from amuse.support.interface import InCodeComponentImplementation

from amuse.test.amusetest import TestWithMPI
from amuse.test import compile_tools
import os
import time
import shlex

from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.rfi import channel
from amuse.rfi.core import *

from .test_fortran_implementation import codestring, ForTestingInterface


class ForTesting(InCodeComponentImplementation):

    def __init__(self, exefile, **options):
        InCodeComponentImplementation.__init__(self, ForTestingInterface(exefile, **options), **options)


class TestInterface(TestWithMPI):

    @classmethod
    def setup_class(cls):
        print("building")
        cls.check_can_compile_modules()
        cls.check_fortran_version()
        cls.check_not_in_mpiexec()
        cls.exefile = compile_tools.build_fortran_worker(codestring, cls.get_path_to_results(), ForTestingInterface)
        print("done")

    @classmethod
    def check_fortran_version(self):
        pass

    @classmethod
    def check_not_in_mpiexec(cls):
        """
        The tests will fork another process, if the test run
        is itself an mpi process, the tests may fail. 

        For the hydra process manager the tests will fail.
        So skip the tests if we detect hydra
        """
        if 'HYDI_CONTROL_FD' in os.environ:
            return  # can run in modern mpiexec.hydra
        if 'HYDRA_CONTROL_FD' in os.environ or 'PMI_FD' in os.environ:
            cls.skip('cannot run the socket tests under hydra process manager')

    def test1(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        int_out, error = instance.echo_int(10)
        del instance
        self.assertEqual(int_out, 10)
        self.assertEqual(error, 0)

    def test2(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        out, error = instance.echo_double(4.0)
        del instance
        self.assertEqual(out, 4.0)
        self.assertEqual(error, 0)

    def test3(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        input = [1, 2, 3, 4]
        output, errors = instance.echo_int(input)
        del instance
        self.assertEqual(len(errors), 4)
        for actual, expected in zip(output, input):
            self.assertEqual(actual, expected)

    def test4(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        input = [1.0, 2.1, 3.3, 4.2]
        output, errors = instance.echo_double(input)
        del instance
        self.assertEqual(len(errors), 4)
        for actual, expected in zip(output, input):
            self.assertEqual(actual, expected)

    def test5(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        out, error = instance.echo_float(4.0)
        del instance
        self.assertEqual(out, 4.0)
        self.assertEqual(error, 0)

    def test6(self):

        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        out, error = instance.echo_string("abc")
        del instance
        self.assertEqual(error, 0)
        self.assertEqual(out, "abc")

    def test7(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        out, error = instance.echo_string(["abc", "def"])
        del instance

        self.assertEqual(error[0], 0)
        self.assertEqual(error[1], 0)
        self.assertEqual(out[0], "abc")
        self.assertEqual(out[1], "def")

    def test8(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        out1, out2, error = instance.echo_strings("abc", "def")
        del instance

        self.assertEqual(error, 0)
        self.assertEqual(out1, "Abc")
        self.assertEqual(out2, "Bef")

    def test9(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        str1_out, str2_out, error = instance.echo_strings(["abc", "def"], ["ghi", "jkl"])
        del instance

        self.assertEqual(error[0], 0)
        self.assertEqual(error[1], 0)
        self.assertEqual(str1_out[0], "Abc")
        self.assertEqual(str1_out[1], "Aef")
        self.assertEqual(str2_out[0], "Bhi")
        self.assertEqual(str2_out[1], "Bkl")

    def test10(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        out = instance.return_string("qwerty")
        out = instance.return_string("abcdefghi")

        instance.stop()
        del instance

        self.assertEqual(out, "abcdefghi")

    def test11(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        out, error = instance.hello_string()
        del instance

        self.assertEqual(out, "hello")

    def test12(self):

        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        out, error = instance.echo_string_fixed_len("abc")
        del instance
        self.assertEqual(error, 0)
        self.assertEqual(out, "abc")

    def test13(self):
        instance = ForTestingInterface(self.exefile, debugger="none", channel_type="sockets")
        (output_ints, error) = instance.echo_array_with_result([4, 5, 6])
        instance.stop()
        print(output_ints, error)
        self.assertEqual(output_ints[0], 4)
        self.assertEqual(output_ints[1], 5)
        self.assertEqual(output_ints[2], 6)

        self.assertEqual(error[0], -1)
        self.assertEqual(error[1], -1)
        self.assertEqual(error[2], -1)

    def test14(self):

        for x in range(4):
            instance = ForTestingInterface(self.exefile, channel_type="sockets")
            int_out, error = instance.echo_int(10)
            instance.stop()
            self.assertEqual(int_out, 10)
            self.assertEqual(error, 0)

    def test15(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        (output_ints, error) = instance.echo_inout_array_with_result([4, 5, 6])
        instance.stop()
        self.assertEqual(output_ints[0], 14)
        self.assertEqual(output_ints[1], 15)
        self.assertEqual(output_ints[2], 16)

        self.assertEqual(error[0], 11)
        self.assertEqual(error[1], 11)
        self.assertEqual(error[2], 11)

    def test16(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        (output1, error1) = instance.echo_logical(True)
        (output2, error2) = instance.echo_logical(False)
        instance.stop()
        self.assertEqual(error1, 0)
        self.assertEqual(error2, 0)
        self.assertTrue(output1)
        self.assertFalse(output2)

    def test16b(self):
        instance = ForTesting(self.exefile, channel_type="sockets")
        output = instance.echo_logical([True, True, False, True, False])
        self.assertEqual(output, [True, True, False, True, False])

    def test16c(self):
        instance = ForTesting(self.exefile, redirection="none")
        output = instance.echo_logical2([True, True, False, True, False]*1024)
        self.assertEqual(output, [True, True, False, True, False]*1024)

    def xtest20(self):
        #
        # TURNED OFF support for redirection,
        # by default output is redirected to /dev/null
        # if you need file, use the support from your mpi implementation
        #
        if os.path.exists("pout.000"):
            os.remove("pout.000")
        if os.path.exists("perr.000"):
            os.remove("perr.000")

        x = ForTestingInterface(self.exefile, redirect_stderr_file='perr', redirect_stdout_file='pout', redirection="file", channel_type="sockets")
        x.print_string("abc")
        x.print_error_string("exex")
        x.stop()

        time.sleep(0.2)

        self.assertTrue(os.path.exists("pout.000"))
        with open("pout.000", "r") as f:
            content = f.read()
        self.assertEqual(content.strip(), "abc")

        self.assertTrue(os.path.exists("perr.000"))
        with open("perr.000", "r") as f:
            content = f.read()
        self.assertEqual(content.strip(), "exex")

        x = ForTestingInterface(self.exefile, redirect_stderr_file='perr', redirect_stdout_file='pout', redirection="file", channel_type="sockets")
        x.print_string("def")
        x.print_error_string("exex")
        x.stop()

        time.sleep(0.2)

        self.assertTrue(os.path.exists("pout.000"))
        with open("pout.000", "r") as f:
            content = f.read()
        self.assertEqual(content.strip(), "abc\n def")

        self.assertTrue(os.path.exists("perr.000"))
        with open("perr.000", "r") as f:
            content = f.read()
        self.assertEqual(content.strip(), "exex\n exex")

    def test35(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        out, error = instance.echo_string(["abc", "def"]*100000)
        del instance

        self.assertEqual(error[0], 0)
        self.assertEqual(error[1], 0)
        self.assertEqual(out[-2], "abc")
        self.assertEqual(out[-1], "def")

    def test36(self):
        instance = ForTestingInterface(self.exefile, channel_type="sockets")
        N = 255
        out, error = instance.echo_string("a"*N)
        del instance

        self.assertEqual(error, 0)
        self.assertEqual(out, "a"*N)



from amuse.test import amusetest

from amuse.units import units
from amuse.units import constants
from amuse.units import nbody_system
import numpy
import time
import sys

from amuse.datamodel import Particles
from amuse.datamodel import Particle
from amuse.couple import encounters


# codes to use
from amuse.community.kepler.interface import Kepler
from amuse.community.smalln.interface import SmallN


def new_binary(
        mass1, mass2, semi_major_axis,
        eccentricity=0, keyoffset=1,
        is_at_periapsis=True,
        G=nbody_system.G
    ):
    total_mass = mass1 + mass2
    mass_fraction_particle_1 = mass1 / (total_mass)

    binary = Particles(keys=range(keyoffset, keyoffset+2))
    binary[0].mass = mass1
    binary[1].mass = mass2
    binary.child1 = None
    binary.child2 = None

    mu = G * total_mass

    if is_at_periapsis:
        velocity = numpy.sqrt(mu / semi_major_axis * ((1.0 + eccentricity)/(1.0 - eccentricity)))
        radius = semi_major_axis * (1.0 - eccentricity)
    else:
        velocity = numpy.sqrt(mu / semi_major_axis * ((1.0 - eccentricity)/(1.0 + eccentricity)))
        radius = semi_major_axis * (1.0 + eccentricity)

    binary[0].position = ((1.0 - mass_fraction_particle_1) * radius * [1.0, 0.0, 0.0])
    binary[1].position = -(mass_fraction_particle_1 * radius * [1.0, 0.0, 0.0])

    binary[0].velocity = ((1.0 - mass_fraction_particle_1) * velocity * [0.0, 1.0, 0.0])
    binary[1].velocity = -(mass_fraction_particle_1 * velocity * [0.0, 1.0, 0.0])

    return binary


class TryHandleEncounter(encounters.AbstractHandleEncounter, encounters.SelectNeighboursByDistanceMixin):

    def __init__(self,
        kepler_code,
        G=nbody_system.G
    ):
        encounters.AbstractHandleEncounter.__init__(
            self,
            kepler_code,
            G
        )
        encounters.SelectNeighboursByDistanceMixin.__init__(self)


class TestAbstractHandleEncounter(amusetest.TestWithMPI):

    def new_kepler(self):
        x = Kepler()
        x.initialize_code()
        return x

    def test1(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.5, 0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length

        particles_in_field = Particles(5)
        particles_in_field.mass = [1, 2, 3, 4, 5] | nbody_system.mass
        particles_in_field[0].position = [2, 0, 0] | nbody_system.length
        particles_in_field[1].position = [1.5, 0, 0] | nbody_system.length
        particles_in_field[2].position = [0.5, 1, 0] | nbody_system.length
        particles_in_field[3].position = [0.5, -0.5, 0] | nbody_system.length
        particles_in_field[4].position = [0, 0, 2] | nbody_system.length
        particles_in_field.velocity = [0, 0.0, 0] | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)

        x.execute()

        self.assertAlmostRelativeEqual(x.large_scale_of_particles_in_the_encounter, 1.0 | nbody_system.length)
        self.assertAlmostRelativeEqual(x.small_scale_of_particles_in_the_encounter, 3.0 | nbody_system.length)

        self.assertEqual(len(x.particles_close_to_encounter), 3)
        self.assertAlmostRelativeEqual(x.particles_close_to_encounter.mass, [2, 3, 4] | nbody_system.mass)

    def test2(self):

        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.5, 0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length

        particles_in_field = Particles(keys=(5,))
        particles_in_field.mass = 2 | nbody_system.mass
        particles_in_field.position = [-0.5, 0, 0] | nbody_system.length
        particles_in_field.velocity = [0, 0.5, 0] | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length

        particles_in_multiples = Particles()
        particles_in_multiples.add_particle(particles_in_encounter[0])
        multiple = particles_in_multiples[0]
        multiple.components = Particles(keys=(3, 4))
        multiple.components.mass = 0.5 | nbody_system.mass
        multiple.components[0].position = [0, 0.1, 0] | nbody_system.length
        multiple.components[1].position = [0, -0.1, 0] | nbody_system.length
        multiple.components[0].velocity = [0, 0, 0.2] | nbody_system.speed
        multiple.components[1].velocity = [0, 0.1, -0.2] | nbody_system.speed
        multiple.components.child1 = None
        multiple.components.child2 = None

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)
        x.existing_multiples.add_particles(particles_in_multiples)

        x.execute()

        self.assertEqual(len(x.all_singles_in_encounter), 3)
        self.assertEqual(len(x.all_singles_in_evolve), 4)

        self.assertAlmostRelativeEqual(x.all_singles_in_evolve.mass, [0.5, 0.5, 1.0, 2.0] | nbody_system.mass)

        self.assertEqual(len(x.released_singles), 2)
        print(x.particles_after_encounter)
        child1 = x.particles_after_encounter[-2]
        child2 = x.particles_after_encounter[-1]
        print(child1.position)
        print(child2.position)
        self.assertAlmostRelativeEqual(child1.position, [1, 0.1, 0] | nbody_system.length)
        self.assertAlmostRelativeEqual(child1.velocity, [0, 0.5, 0.2] | nbody_system.speed)
        self.assertAlmostRelativeEqual(child2.position, [1, -0.1, 0] | nbody_system.length)
        self.assertAlmostRelativeEqual(child2.velocity, [0, 0.6, -0.2] | nbody_system.speed)

    def test3(self):

        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.0, 0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length

        particles_in_field = Particles(keys=(2, 4, 5, 6, 7,))
        particles_in_field.mass = [1, 2, 3, 4, 5] | nbody_system.mass
        particles_in_field[0].position = [2, 0, 0] | nbody_system.length
        particles_in_field[1].position = [1.5, 0, 0] | nbody_system.length
        particles_in_field[2].position = [0.5, 1, 0] | nbody_system.length
        particles_in_field[3].position = [0.5, -0.5, 0] | nbody_system.length
        particles_in_field[4].position = [0, 0, 2] | nbody_system.length
        particles_in_field.velocity = [0, 0, 0] | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)

        simple_binary = new_binary(
            1 | nbody_system.mass,
            1 | nbody_system.mass,
            0.2 | nbody_system.length
        )

        def evolve_singles_in_encounter_until_end_state():
            particles = x.singles_and_multiples_after_evolve
            particles.add_particles(x.all_singles_in_encounter)
            particles.child1 = None
            particles.child2 = None

            root_particle = particles.add_particle(Particle(
                key=10,
                mass=2.0 | nbody_system.mass,
                position=particles[0:2].center_of_mass(),
                velocity=particles[0:2].center_of_mass_velocity(),
            ))
            root_particle.child1 = particles[0]
            root_particle.child2 = particles[1]
            particles[0].position = simple_binary[0].position + root_particle.position
            particles[1].position = simple_binary[1].position + root_particle.position

            particles[0].velocity = simple_binary[0].velocity + root_particle.velocity
            particles[1].velocity = simple_binary[1].velocity + root_particle.velocity

        x.evolve_singles_in_encounter_until_end_state = evolve_singles_in_encounter_until_end_state
        x.determine_structure_of_the_evolved_state = lambda: 1

        x.execute()
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)

        self.assertAlmostRelativeEqual(multiple.components[0].position, simple_binary[0].position)
        self.assertAlmostRelativeEqual(multiple.components[1].position, simple_binary[1].position)

    def test4(self):

        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.0, 0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length

        particles_in_field = Particles(keys=(3, 4, 5, 6, 7,))
        particles_in_field.mass = [1, 2, 3, 4, 5] | nbody_system.mass
        particles_in_field[0].position = [2, 0, 0] | nbody_system.length
        particles_in_field[1].position = [1.5, 0, 0] | nbody_system.length
        particles_in_field[2].position = [0.5, 1, 0] | nbody_system.length
        particles_in_field[3].position = [0.5, -0.5, 0] | nbody_system.length
        particles_in_field[4].position = [0, 0, 2] | nbody_system.length
        particles_in_field.velocity = [0, 0, 0] | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length

        binaries = Particles(keys=(20,))
        binaries[0].child1 = particles_in_encounter[0]
        binaries[0].child2 = particles_in_encounter[1]

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)
        x.existing_binaries.add_particles(binaries)

        simple_binary = new_binary(
            1 | nbody_system.mass,
            1 | nbody_system.mass,
            0.2 | nbody_system.length
        )

        def evolve_singles_in_encounter_until_end_state():
            particles = x.singles_and_multiples_after_evolve
            particles.add_particles(x.all_singles_in_encounter)
            particles.child1 = None
            particles.child2 = None

            root_particle = particles.add_particle(Particle(
                key=10,
                mass=2.0 | nbody_system.mass,
                position=particles[0:2].center_of_mass(),
                velocity=particles[0:2].center_of_mass_velocity(),
            ))
            root_particle.child1 = particles[0]
            root_particle.child2 = particles[1]
            particles[0].position = simple_binary[0].position + root_particle.position
            particles[1].position = simple_binary[1].position + root_particle.position

            particles[0].velocity = simple_binary[0].velocity + root_particle.velocity
            particles[1].velocity = simple_binary[1].velocity + root_particle.velocity

        x.evolve_singles_in_encounter_until_end_state = evolve_singles_in_encounter_until_end_state
        x.determine_structure_of_the_evolved_state = lambda: 1

        x.execute()
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 0)
        self.assertEqual(len(x.updated_binaries), 1)
        self.assertEqual(x.updated_binaries[0], binaries[0])

    def test5(self):

        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.0, 0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length

        particles_in_field = Particles(keys=(2, 4, 5, 6, 7,))
        particles_in_field.mass = [1, 2, 3, 4, 5] | nbody_system.mass
        particles_in_field[0].position = [2, 0, 0] | nbody_system.length
        particles_in_field[1].position = [1.5, 0, 0] | nbody_system.length
        particles_in_field[2].position = [0.5, 1, 0] | nbody_system.length
        particles_in_field[3].position = [0.5, -0.5, 0] | nbody_system.length
        particles_in_field[4].position = [0, 0, 2] | nbody_system.length
        particles_in_field.velocity = [0, 0, 0] | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.parameters.hard_binary_factor = 1
        x.small_scale_factor = 1

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)

        simple_binary = new_binary(
            1 | nbody_system.mass,
            1 | nbody_system.mass,
            2 | nbody_system.length
        )

        def evolve_singles_in_encounter_until_end_state():
            particles = x.singles_and_multiples_after_evolve
            particles.add_particles(x.all_singles_in_encounter)
            particles.child1 = None
            particles.child2 = None

            root_particle = particles.add_particle(Particle(
                key=10,
                mass=2.0 | nbody_system.mass,
                position=particles[0:2].center_of_mass(),
                velocity=particles[0:2].center_of_mass_velocity(),
            ))
            root_particle.child1 = particles[0]
            root_particle.child2 = particles[1]
            particles[0].position = simple_binary[0].position + root_particle.position
            particles[1].position = simple_binary[1].position + root_particle.position

            particles[0].velocity = simple_binary[0].velocity + root_particle.velocity
            particles[1].velocity = simple_binary[1].velocity + root_particle.velocity

        x.evolve_singles_in_encounter_until_end_state = evolve_singles_in_encounter_until_end_state
        x.determine_structure_of_the_evolved_state = lambda: 1

        x.execute()

        # no multiples as the binary is larger than the
        # hard binary scale
        self.assertEqual(len(x.new_multiples), 0)
        self.assertEqual(len(x.new_binaries), 0)

    def test6(self):

        particles_in_encounter = Particles(keys=(1, 2, 3))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter[2].position = [0, 0.5, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.0, 0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.small_scale_factor = 1.0

        simple_binary_1 = new_binary(
            1 | nbody_system.mass,
            1 | nbody_system.mass,
            0.4 | nbody_system.length
        )
        simple_binary_top = new_binary(
            2 | nbody_system.mass,
            1 | nbody_system.mass,
            2 | nbody_system.length
        )

        def evolve_singles_in_encounter_until_end_state():
            particles = x.singles_and_multiples_after_evolve
            particles.add_particles(x.all_singles_in_encounter)
            particles.child1 = None
            particles.child2 = None

            inner_binary_particle = particles.add_particle(Particle(
                key=10,
                mass=2.0 | nbody_system.mass,
                position=particles[0:2].center_of_mass(),
                velocity=particles[0:2].center_of_mass_velocity(),
            ))
            inner_binary_particle.child1 = particles[0]
            inner_binary_particle.child2 = particles[1]
            particles[0].position = simple_binary_1[0].position + inner_binary_particle.position
            particles[1].position = simple_binary_1[1].position + inner_binary_particle.position

            particles[0].velocity = simple_binary_1[0].velocity + inner_binary_particle.velocity
            particles[1].velocity = simple_binary_1[1].velocity + inner_binary_particle.velocity

            root_particle = particles.add_particle(Particle(
                key=11,
                mass=3.0 | nbody_system.mass,
                position=particles.center_of_mass(),
                velocity=particles.center_of_mass_velocity(),
            ))

            root_particle.child1 = inner_binary_particle
            root_particle.child2 = particles[2]
            inner_binary_particle.position = simple_binary_top[0].position + root_particle.position
            particles[2].position = simple_binary_top[1].position + root_particle.position

            inner_binary_particle.velocity = simple_binary_top[0].velocity + root_particle.velocity
            particles[2].velocity = simple_binary_top[1].velocity + root_particle.velocity

        x.evolve_singles_in_encounter_until_end_state = evolve_singles_in_encounter_until_end_state
        x.determine_structure_of_the_evolved_state = lambda: 1

        x.execute()

        # no multiples as the binary is larger than the
        # hard binary scale
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)
        self.assertAlmostRelativeEqual(multiple.components[0].key, particles_in_encounter[0].key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, particles_in_encounter[1].key)

    def test7(self):

        particles_in_encounter = Particles(keys=(1, 2, 3))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter[2].position = [0, 0.5, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.0, 0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )
        x.particles_in_encounter.add_particles(particles_in_encounter)

        multiples = Particles()
        particle_in_multiples = multiples.add_particle(particles_in_encounter[2])
        particle_in_multiples.components = Particles(keys=(4, 5))
        particle_in_multiples.components.mass = 1 | nbody_system.mass
        particle_in_multiples.components[0].position = [0, 0, 0.2] | nbody_system.length
        particle_in_multiples.components[1].position = [0, 0, -0.2] | nbody_system.length
        particle_in_multiples.components.velocity = [0, 0.0, 0] | nbody_system.speed
        particle_in_multiples.components.child1 = None
        particle_in_multiples.components.child2 = None
        x.existing_multiples.add_particles(multiples)

        def evolve_singles_in_encounter_until_end_state():

            particles = x.singles_and_multiples_after_evolve
            particles.add_particles(x.all_singles_in_encounter)
            print(particles)
            particles.child1 = None
            particles.child2 = None

        x.evolve_singles_in_encounter_until_end_state = evolve_singles_in_encounter_until_end_state
        x.determine_structure_of_the_evolved_state = lambda: 1

        x.execute()

        # no multiples as the binary is larger than the
        # hard binary scale
        self.assertEqual(len(x.new_multiples), 0)
        self.assertEqual(len(x.new_binaries), 0)
        self.assertEqual(len(x.dissolved_multiples), 1)
        self.assertEqual(len(x.released_singles), 2)
        self.assertTrue(particle_in_multiples.components[0] in x.released_singles)
        self.assertTrue(particle_in_multiples.components[1] in x.released_singles)

    def xtest8(self):
        particles_in_encounter = Particles(3)
        particles_in_encounter.mass = 1. | nbody_system.mass
        particles_in_encounter[0].position = [0., 0., 0.] | nbody_system.length
        particles_in_encounter[1].position = [1., 0., 0.] | nbody_system.length
        particles_in_encounter[2].position = [2., 0., 0.] | nbody_system.length
        particles_in_encounter.velocity = [0., 0.5, 0.] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 1., 0.2] | nbody_system.length

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)

        x.execute()

        self.assertAlmostRelativeEqual(x.large_scale_of_particles_in_the_encounter, 2.0 | nbody_system.length)
        self.assertAlmostRelativeEqual(x.small_scale_of_particles_in_the_encounter, 1.5 | nbody_system.length)

    def test8b(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.5, 0] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 0.5] | nbody_system.length

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)

        x.execute()

        self.assertAlmostRelativeEqual(x.initial_potential_in_field,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_energy, -0.75 | nbody_system.energy)

    def test9(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.5, 0] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 0.5] | nbody_system.length

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)

        x.execute()

        self.assertAlmostRelativeEqual(x.initial_potential_in_field,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_singles_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.delta_phi_1,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.delta_phi_2, 0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_multiple_energy,  0 | nbody_system.energy)

    def test10(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.5, 0] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 0.5] | nbody_system.length

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particle(Particle(
            mass=1 | nbody_system.mass,
            position=[5, 0, 0] | nbody_system.length,
            velocity=[0, 0, 0] | nbody_system.speed,
            radius=0 | nbody_system.length
        ))

        x.execute()

        self.assertAlmostRelativeEqual(x.initial_potential_in_field,   -0.45 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_singles_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.delta_phi_1,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.delta_phi_2, 0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_multiple_energy,  0 | nbody_system.energy)

    def test11(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.5, 0] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 0.5] | nbody_system.length

        particles_in_multiples = Particles()
        particles_in_multiples.add_particle(particles_in_encounter[0])
        multiple = particles_in_multiples[0]
        multiple.components = Particles(keys=(3, 4))
        multiple.components.mass = 0.5 | nbody_system.mass
        multiple.components[0].position = [0, 0.1, 0] | nbody_system.length
        multiple.components[1].position = [0, -0.1, 0] | nbody_system.length
        multiple.components[0].velocity = [0, 0, 0.2] | nbody_system.speed
        multiple.components[1].velocity = [0, 0.1, -0.2] | nbody_system.speed
        multiple.components.child1 = None
        multiple.components.child2 = None

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particle(Particle(
            mass=1 | nbody_system.mass,
            position=[5, 0, 0] | nbody_system.length,
            velocity=[0, 0, 0] | nbody_system.speed,
            radius=0 | nbody_system.length
        ))
        x.existing_multiples.add_particles(particles_in_multiples)

        x.execute()

        self.assertAlmostRelativeEqual(x.initial_potential_in_field,   -0.45 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_singles_energy, -1.912495 | nbody_system.energy, 4)
        self.assertAlmostRelativeEqual(x.delta_phi_1,  0.06500400 | nbody_system.energy, 4)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy,  -1.2275 | nbody_system.energy, 4)
        self.assertAlmostRelativeEqual(x.final_energy, -1.9124959996 | nbody_system.energy, 4)
        self.assertAlmostRelativeEqual(x.delta_phi_2, 0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_multiple_energy,  0 | nbody_system.energy)

    def test12(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.5, 0] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 0.5] | nbody_system.length

        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particle(Particle(
            mass=0.1 | nbody_system.mass,
            position=[1.5, 0, 0] | nbody_system.length,
            velocity=[0, 0, 0] | nbody_system.speed,
            radius=1 | nbody_system.length
        ))

        x.execute()

        self.assertAlmostRelativeEqual(x.initial_potential_in_field,   0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_energy, -1.016666 | nbody_system.energy, 4)
        self.assertAlmostRelativeEqual(x.initial_singles_energy, -1.016666 | nbody_system.energy, 4)
        self.assertAlmostRelativeEqual(x.delta_phi_1,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy,  0 | nbody_system.energy)

    def test13(self):
        x = TryHandleEncounter(
            G=nbody_system.G,
            kepler_code=self.new_kepler()
        )

        x.parameters.hard_binary_factor = 1
        self.assertEqual(x.parameters.hard_binary_factor, 1)
        self.assertEqual(x.parameters.scatter_factor, 10)
        x.parameters.scatter_factor = 2
        self.assertEqual(x.parameters.scatter_factor, 2)
        self.assertEqual(x.get_scatter_factor(), 2)


class TestKeplerOrbits(amusetest.TestWithMPI):

    def new_kepler(self):
        x = Kepler()
        x.initialize_code()
        return x

    def test1(self):
        x = encounters.KeplerOrbits(self.new_kepler())
        binary = new_binary(
            1 | nbody_system.mass,
            0.5 | nbody_system.mass,
            1.2 | nbody_system.length,
        )
        semimajor_axis, eccentricity = x.get_semimajor_axis_and_eccentricity_for_binary_components(
            binary[0],
            binary[1]
        )
        self.assertAlmostRelativeEquals(semimajor_axis,  1.2 | nbody_system.length)
        self.assertAlmostRelativeEquals(eccentricity,  0)

    def test2(self):
        x = encounters.KeplerOrbits(self.new_kepler())
        binary = new_binary(
            1 | nbody_system.mass,
            0.5 | nbody_system.mass,
            2.0 | nbody_system.length,
            0.5,
            is_at_periapsis=False

        )
        binary.position += [0.1, 0.2, 0.3] | nbody_system.length
        binary.velocity += [0.4, 0.5, 0.6] | nbody_system.speed

        dpos, dvel = x.compress_binary(
            binary,
            1.2 | nbody_system.length
        )

        center_of_mass_before = binary.center_of_mass()
        center_of_mass_velocity_before = binary.center_of_mass_velocity()

        binary.position += dpos
        binary.velocity += dvel

        center_of_mass_after = binary.center_of_mass()
        center_of_mass_velocity_after = binary.center_of_mass_velocity()

        self.assertAlmostRelativeEquals(center_of_mass_before, center_of_mass_after)
        self.assertAlmostRelativeEquals(center_of_mass_velocity_before, center_of_mass_velocity_after)
        separation = (binary[0].position - binary[1].position).length()
        self.assertAlmostRelativeEquals(separation, 1.2 | nbody_system.length)

    def test3(self):
        x = encounters.KeplerOrbits(self.new_kepler())
        binary = new_binary(
            1 | nbody_system.mass,
            0.5 | nbody_system.mass,
            2.0 | nbody_system.length,
            0.5,
            is_at_periapsis=True

        )
        binary.position += [0.1, 0.2, 0.3] | nbody_system.length
        binary.velocity += [0.4, 0.5, 0.6] | nbody_system.speed

        dpos, dvel = x.expand_binary(
            binary,
            2.5 | nbody_system.length
        )

        center_of_mass_before = binary.center_of_mass()
        center_of_mass_velocity_before = binary.center_of_mass_velocity()

        binary.position += dpos
        binary.velocity += dvel

        center_of_mass_after = binary.center_of_mass()
        center_of_mass_velocity_after = binary.center_of_mass_velocity()

        self.assertAlmostRelativeEquals(center_of_mass_before, center_of_mass_after, 8)
        self.assertAlmostRelativeEquals(center_of_mass_velocity_before, center_of_mass_velocity_after, 8)
        separation = (binary[0].position - binary[1].position).length()
        self.assertAlmostRelativeEquals(separation, 2.5 | nbody_system.length)

    def test4(self):
        converter = nbody_system.nbody_to_si(1 | units.MSun, 1 | units.AU)
        kepler = Kepler(converter)
        kepler.initialize_code()
        x = encounters.KeplerOrbits(kepler)
        binary = new_binary(
            1 | units.MSun,
            0.5 | units.MSun,
            1.2 | units.AU,
            G=constants.G
        )
        semimajor_axis, eccentricity = x.get_semimajor_axis_and_eccentricity_for_binary_components(
            binary[0],
            binary[1]
        )
        self.assertAlmostRelativeEquals(semimajor_axis,  1.2 | units.AU, 8)
        self.assertAlmostRelativeEquals(eccentricity,  0)


class TestScaleSystem(amusetest.TestWithMPI):

    def new_kepler(self, converter=None):
        x = Kepler(converter)
        x.initialize_code()
        return x

    def test1(self):
        kepler = encounters.KeplerOrbits(self.new_kepler())
        binary = new_binary(
            1 | nbody_system.mass,
            0.5 | nbody_system.mass,
            1.2 | nbody_system.length,
            0.5,
            is_at_periapsis=False
        )
        binary.radius = 0 | nbody_system.length

        x = encounters.ScaleSystem(kepler)
        self.assertTrue((binary[0].position - binary[1].position).length() > 1.5 | nbody_system.length)
        x.scale_particles_to_sphere(binary, 0.75 | nbody_system.length)
        self.assertTrue((binary[0].position - binary[1].position).length() <= 1.6 | nbody_system.length)
        print((binary[0].position - binary[1].position).length())
        self.assertTrue((binary[0].position - binary[1].position).length() >= (1.5 - 1e-6) | nbody_system.length)

    def test2(self):
        kepler = encounters.KeplerOrbits(self.new_kepler())
        binary = new_binary(
            1 | nbody_system.mass,
            0.5 | nbody_system.mass,
            1.2 | nbody_system.length,
            0.5,
            is_at_periapsis=True
        )
        binary.radius = 0 | nbody_system.length

        x = encounters.ScaleSystem(kepler)
        self.assertTrue((binary[0].position - binary[1].position).length() < 1.5 | nbody_system.length)
        x.scale_particles_to_sphere(binary, 0.75 | nbody_system.length)
        self.assertTrue((binary[0].position - binary[1].position).length() <= 1.6 | nbody_system.length)
        print((binary[0].position - binary[1].position).length())
        self.assertTrue((binary[0].position - binary[1].position).length() >= (1.5 - 1e-6) | nbody_system.length)

    def test3(self):
        kepler = encounters.KeplerOrbits(self.new_kepler())

        particles = Particles(keys=(1, 2))
        particles.mass = 1 | nbody_system.mass
        particles.child2 = None
        particles.child1 = None
        particles[0].position = [1, 0, 0] | nbody_system.length
        particles[0].velocity = [1, 0.0, 0] | nbody_system.speed

        particles[1].position = [-1, 0, 0] | nbody_system.length
        particles[1].velocity = [-1, 0.0, 0] | nbody_system.speed

        particles.radius = 0 | nbody_system.length

        x = encounters.ScaleSystem(kepler)

        print((particles[0].position - particles[1].position).length())
        self.assertTrue((particles[0].position - particles[1].position).length() > 1.5 | nbody_system.length)

        x.scale_particles_to_sphere(particles, 0.75 | nbody_system.length)

        self.assertTrue((particles[0].position - particles[1].position).length() <= 1.6 | nbody_system.length)
        print(particles)
        self.assertTrue((particles[0].position - particles[1].position).length() >= (1.5 - 1e-6) | nbody_system.length)
        self.assertAlmostRelativeEquals(particles[0].position,  [0.75, 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(particles[1].position,  [-0.75, 0, 0] | nbody_system.length)

    def test4(self):
        kepler = encounters.KeplerOrbits(self.new_kepler())

        particles = Particles(keys=(1, 2, 3, 4, 5, 6))
        particles.child2 = None
        particles.child1 = None
        particles.mass = 1 | nbody_system.mass
        for i in range(3):
            position = [0, 0, 0] | nbody_system.length
            position[i] = 1 | nbody_system.length
            particles[i].position = position
            position[i] = -1 | nbody_system.length
            particles[i+3].position = position

            velocity = [0, 0, 0] | nbody_system.speed
            velocity[i] = 1 | nbody_system.speed
            particles[i].velocity = velocity
            velocity[i] = -1 | nbody_system.speed
            particles[i+3].velocity = velocity

        particles.radius = 0 | nbody_system.length

        x = encounters.ScaleSystem(kepler)
        potential_energy0 = particles.potential_energy(G=nbody_system.G)
        kinetic_energy0 = particles.kinetic_energy()

        x.scale_particles_to_sphere(particles, 0.5 | nbody_system.length)

        self.assertAlmostRelativeEquals(particles[0].position,  [1/numpy.sqrt(2), 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(particles[3].position,  [-1/numpy.sqrt(2), 0, 0] | nbody_system.length)
        self.assertAlmostRelativeEquals(particles[0].velocity,  [1.542, 0, 0] | nbody_system.speed, 3)
        self.assertAlmostRelativeEquals(particles[3].velocity,  [-1.542, 0, 0] | nbody_system.speed, 3)

        potential_energy1 = particles.potential_energy(G=nbody_system.G)
        kinetic_energy1 = particles.kinetic_energy()

        self.assertAlmostRelativeEquals(potential_energy0 + kinetic_energy0, potential_energy1 + kinetic_energy1)

    def test5(self):
        converter = nbody_system.nbody_to_si(1 | units.MSun, 1 | units.AU)
        kepler = encounters.KeplerOrbits(self.new_kepler(converter))

        particles = Particles(keys=(1, 2, 3))
        particles.child2 = None
        particles.child1 = None
        particles.position = [
            [-1.28230200e-05,  -3.69457095e-05,  -2.02383488e-05],
            [-2.91749746e-05,  -1.21387289e-05,   1.56377986e-07],
            [2.92123436e-05,   1.22463965e-05,  -9.73992061e-08]
        ] | units.parsec
        particles.velocity = [
            [-8685.98414414, -413260.30543051, -135268.19175611],
            [19639.55455979,  -30251.55372943,     972.69648982],
            [-19614.24178601,   31455.88066005,    -578.49669843]
        ] | (units.m / units.s)
        particles.mass = [0.14571045,  50., 50.] | units.MSun
        particles.radius = [8, 0., 0.0] | units.AU

        x = encounters.ScaleSystem(kepler, G=constants.G)

        potential_energy0 = particles.potential_energy()
        kinetic_energy0 = particles.kinetic_energy()

        x.scale_particles_to_sphere(particles, 4 | units.AU)

        potential_energy1 = particles.potential_energy()
        kinetic_energy1 = particles.kinetic_energy()

        self.assertAlmostRelativeEquals(potential_energy0 + kinetic_energy0, potential_energy1 + kinetic_energy1)


class TestHandleEncounter(amusetest.TestWithMPI):

    def new_kepler(self):
        x = Kepler()
        x.initialize_code()
        return x

    def test1(self):
        particles_in_encounter = Particles(keys=(1, 2, 3))
        particles_in_encounter.mass = 1. | nbody_system.mass
        particles_in_encounter[0].position = [1., 0., 0.] | nbody_system.length
        particles_in_encounter[1].position = [0., 0., 0.] | nbody_system.length
        particles_in_encounter[2].position = [0., 0.5, 0.] | nbody_system.length
        particles_in_encounter.velocity = [0., 0.0, 0.] | nbody_system.speed
        particles_in_encounter[2].velocity = [0., 0.0, 3.] | nbody_system.speed
        particles_in_encounter.radius = 0. | nbody_system.length

        particles_in_field = Particles()

        x = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=SmallN(),
            interaction_over_code=None,
            G=nbody_system.G
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)

        x.execute()
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)
        self.assertAlmostRelativeEqual(multiple.components[0].key, particles_in_encounter[0].key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, particles_in_encounter[1].key)
        self.assertEqual(len(x.captured_singles), 2)
        self.assertEqual(x.captured_singles.key, [1, 2])

    def test2(self):
        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 2. | nbody_system.mass
        particles_in_encounter[0].position = [1., 0., 0.] | nbody_system.length
        particles_in_encounter[1].position = [0., 0., 0.] | nbody_system.length
        particles_in_encounter.velocity = [0., 0.0, 0.] | nbody_system.speed
        particles_in_encounter.radius = 0. | nbody_system.length

        binary1 = new_binary(
            1. | nbody_system.mass,
            1. | nbody_system.mass,
            0.01 | nbody_system.length,
            keyoffset=30
        )
        binary2 = new_binary(
            1. | nbody_system.mass,
            1. | nbody_system.mass,
            0.01 | nbody_system.length,
            keyoffset=40
        )
        binaries = Particles(keys=(20, 21))
        binaries[0].child1 = binary1[0]
        binaries[0].child2 = binary1[1]
        binaries[1].child1 = binary2[0]
        binaries[1].child2 = binary2[1]
        binary1.child1 = None
        binary1.child2 = None
        binary2.child1 = None
        binary2.child2 = None

        multiples = Particles()
        multiple = particles_in_encounter[0]
        multiple.components = binary1
        multiples.add_particle(multiple)
        multiple = particles_in_encounter[1]
        multiple.components = binary2
        multiples.add_particle(multiple)

        x = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=SmallN(),
            interaction_over_code=None,
            G=nbody_system.G
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.existing_binaries.add_particles(binaries)
        x.existing_multiples.add_particles(multiples)

        x.execute()
        self.assertEqual(len(x.new_multiples), 2)
        self.assertEqual(len(x.new_binaries), 2)
        self.assertEqual(len(x.captured_singles), 0)
        self.assertEqual(len(x.released_singles), 0)

        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)

        self.assertAlmostRelativeEqual(multiple.components[0].key, binaries[0].child1.key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, binaries[1].child1.key)
        multiple = x.new_multiples[1]
        self.assertEqual(len(multiple.components), 2)
        self.assertAlmostRelativeEqual(multiple.components[0].key, binaries[0].child2.key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, binaries[1].child2.key)

    def test3(self):
        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 2. | nbody_system.mass
        particles_in_encounter[0].position = [1., 0., 0.] | nbody_system.length
        particles_in_encounter[1].position = [0., 0., 0.] | nbody_system.length
        particles_in_encounter.velocity = [0., 0.0, 0.] | nbody_system.speed
        particles_in_encounter.radius = 0. | nbody_system.length

        binary1 = new_binary(
            1. | nbody_system.mass,
            1. | nbody_system.mass,
            0.01 | nbody_system.length,
            keyoffset=30
        )
        binaries = Particles(keys=(20,))
        binaries[0].child1 = binary1[0]
        binaries[0].child2 = binary1[1]
        binary1.child1 = None
        binary1.child2 = None

        multiples = Particles()
        multiple = particles_in_encounter[0]
        multiple.components = binary1
        multiples.add_particle(multiple)

        x = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=SmallN(),
            interaction_over_code=None,
            G=nbody_system.G
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.existing_binaries.add_particles(binaries)
        x.existing_multiples.add_particles(multiples)

        x.execute()

        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.dissolved_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        self.assertEqual(len(x.captured_singles), 1)
        self.assertEqual(len(x.released_singles), 1)

        multiple = x.new_multiples[0]
        print(multiple.child1)
        self.assertEqual(len(multiple.components), 2)
        print(multiple.components)

    def test4(self):
        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 1.0 / 20.0 | nbody_system.mass

        particles_in_encounter[0].position = [0.303976184547589401, 0.273137803329168094, 0.0] | nbody_system.length
        particles_in_encounter[1].position = [0.290167020486133631, 0.273139253307546515, 0.0] | nbody_system.length
        particles_in_encounter[0].velocity = [-2.544712989335638387, -1.224759650260411004, 0.0] | nbody_system.speed
        particles_in_encounter[1].velocity = [0.898326624897966997, 0.870611747842778838, 0.0] | nbody_system.speed
        particles_in_encounter.radius = 0.007 | nbody_system.length

        x = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=SmallN(),
            interaction_over_code=None,
            G=nbody_system.G
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)

        x.execute()

        self.assertEqual(len(x.new_multiples), 0)
        self.assertEqual(len(x.dissolved_multiples), 0)
        self.assertEqual(len(x.new_binaries), 0)
        self.assertEqual(len(x.captured_singles), 0)
        self.assertEqual(len(x.released_singles), 0)
        r_before = (particles_in_encounter[0].position - particles_in_encounter[1].position).length()
        r_after = (x.particles_after_encounter[0].position - x.particles_after_encounter[1].position).length()
        print(r_before, r_after)
        self.assertFalse(r_after > (10 * r_before))

    def test5(self):
        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 1.0 / 20.0 | nbody_system.mass

        particles_in_encounter[0].position = [0.0779377282404, -0.559210143918, 0.0] | nbody_system.length
        particles_in_encounter[1].position = [0.07802860386, -0.561207706614, 0.0] | nbody_system.length
        particles_in_encounter[0].velocity = [-1.79086847491, -4.88551917358, 0.0] | nbody_system.speed
        particles_in_encounter[1].velocity = [2.13208424698, 4.31403500143, 0.0] | nbody_system.speed
        particles_in_encounter.radius = 0.001 | nbody_system.length

        x = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=SmallN(),
            interaction_over_code=None,
            G=nbody_system.G
        )
        x.particles_in_encounter.add_particles(particles_in_encounter)

        x.execute()

        self.assertEqual(len(x.new_multiples), 0)
        self.assertEqual(len(x.dissolved_multiples), 0)
        self.assertEqual(len(x.new_binaries), 0)
        self.assertEqual(len(x.captured_singles), 0)
        self.assertEqual(len(x.released_singles), 0)
        r_before = (particles_in_encounter[0].position - particles_in_encounter[1].position).length()
        r_after = (x.particles_after_encounter[0].position - x.particles_after_encounter[1].position).length()
        print(r_before, r_after)
        self.assertFalse(r_after > (10 * r_before))

    def test6(self):
        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 2 | nbody_system.mass
        particles_in_encounter[0].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.0, 0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length

        particles_in_field = Particles(10)
        particles_in_field.mass = 1 | nbody_system.mass
        for i in range(len(particles_in_field)):
            if i == 5:
                j = len(particles_in_field)
            else:
                j = i
            particles_in_field[i].position = [-10 + (2*j), -10 + (2*j), 0] | nbody_system.length
        particles_in_field.velocity = [0, 0.0, 0] | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length

        x = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=SmallN(),
            interaction_over_code=None,
            G=nbody_system.G
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)

        x.execute()

        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.dissolved_multiples), 0)
        self.assertEqual(len(x.new_binaries), 1)
        self.assertEqual(len(x.captured_singles), 2)
        self.assertEqual(len(x.released_singles), 0)

    def test7(self):
        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 2. | nbody_system.mass
        particles_in_encounter[0].position = [1., 0., 0.] | nbody_system.length
        particles_in_encounter[1].position = [0., 0., 0.] | nbody_system.length
        particles_in_encounter.velocity = [0., 0.0, 0.] | nbody_system.speed
        particles_in_encounter.radius = 0. | nbody_system.length

        binary1 = new_binary(
            1. | nbody_system.mass,
            1. | nbody_system.mass,
            0.01 | nbody_system.length,
            keyoffset=30
        )
        binary_energy = binary1.kinetic_energy() + binary1.potential_energy(G=nbody_system.G)
        self.assertAlmostRelativeEqual(binary_energy, -50 | nbody_system.energy)

        binaries = Particles(keys=(20,))
        binaries[0].child1 = binary1[0]
        binaries[0].child2 = binary1[1]
        binary1.child1 = None
        binary1.child2 = None

        multiples = Particles()
        multiple = particles_in_encounter[0]
        multiple.components = binary1
        multiples.add_particle(multiple)

        x = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=SmallN(),
            interaction_over_code=None,
            G=nbody_system.G
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.existing_binaries.add_particles(binaries)
        x.existing_multiples.add_particles(multiples)

        x.execute()
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        self.assertEqual(len(x.captured_singles), 1)
        self.assertEqual(len(x.released_singles), 1)

        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)

        self.assertAlmostRelativeEqual(multiple.components[0].key, binaries[0].child1.key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, particles_in_encounter[1].key)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy, -50 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_multiple_energy, -85.651753404 | nbody_system.energy, 4)

    def test8(self):
        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter[0].mass = 2. | nbody_system.mass
        particles_in_encounter[1].mass = 0.1 | nbody_system.mass
        particles_in_encounter[0].position = [1., 0., 0.] | nbody_system.length
        particles_in_encounter[1].position = [0., 0., 0.] | nbody_system.length
        particles_in_encounter.velocity = [0., 0.0, 0.] | nbody_system.speed
        particles_in_encounter.radius = 0. | nbody_system.length

        binary1 = new_binary(
            1. | nbody_system.mass,
            1. | nbody_system.mass,
            0.01 | nbody_system.length,
            keyoffset=30
        )

        binary_energy = binary1.kinetic_energy() + binary1.potential_energy(G=nbody_system.G)
        self.assertAlmostRelativeEqual(binary_energy, -50 | nbody_system.energy)

        binaries = Particles(keys=(20,))
        binaries[0].child1 = binary1[0]
        binaries[0].child2 = binary1[1]
        binary1.child1 = None
        binary1.child2 = None

        multiples = Particles()
        multiple = particles_in_encounter[0]
        multiple.components = binary1
        multiples.add_particle(multiple)

        x = encounters.HandleEncounter(
            kepler_code=self.new_kepler(),
            resolve_collision_code=SmallN(),
            interaction_over_code=None,
            G=nbody_system.G
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.existing_binaries.add_particles(binaries)
        x.existing_multiples.add_particles(multiples)

        x.execute()
        self.assertEqual(len(x.new_multiples), 0)
        self.assertEqual(len(x.updated_multiples), 1)
        self.assertEqual(len(x.new_binaries), 0)
        self.assertEqual(len(x.captured_singles), 0)
        self.assertEqual(len(x.released_singles), 0)

        multiple = x.updated_multiples[0]
        self.assertEqual(len(multiple.components), 2)
        self.assertEqual(multiple.key, multiples[0].key)

        self.assertEqual(multiple.components[0].key, binaries[0].child1.key)
        self.assertEqual(multiple.components[1].key, binaries[0].child2.key)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy, -50 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_multiple_energy,  -56.2906032319 | nbody_system.energy, 4)


class TestStickyHandleEncounter(amusetest.TestWithMPI):

    def test1(self):
        particles_in_encounter = Particles(keys=(1, 2))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1, 0, 0] | nbody_system.length
        particles_in_encounter[1].position = [0, 0, 0] | nbody_system.length
        particles_in_encounter.velocity = [0, 0.0, 0] | nbody_system.speed
        particles_in_encounter[0].velocity = [-1, 0.0, 0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length

        particles_in_field = Particles()

        x = encounters.StickyHandleEncounter(
            G=nbody_system.G
        )

        x.particles_in_encounter.add_particles(particles_in_encounter)

        x.execute()
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)
        self.assertAlmostRelativeEqual(multiple.components[0].key, particles_in_encounter[0].key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, particles_in_encounter[1].key)
        self.assertEqual(len(x.captured_singles), 2)
        self.assertEqual(x.captured_singles.key, [1, 2])
        self.assertAlmostRelativeEqual(multiple.velocity, [-0.5, 0.0, 0] | nbody_system.speed)
        self.assertAlmostRelativeEqual(multiple.mass, 2 | nbody_system.mass)
        self.assertAlmostRelativeEqual(multiple.radius, 0.5 | nbody_system.length)
        self.assertAlmostRelativeEqual(multiple.position, [0.5, 0, 0] | nbody_system.length)



import os.path
import numpy

from amuse.test.amusetest import get_path_to_results, TestWithMPI
try:
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
    from amuse.plot import plot, semilogy, xlabel, ylabel, loglog
except ImportError:
    HAS_MATPLOTLIB = False

from amuse.support.exceptions import AmuseException, AmuseWarning
from amuse.community.mesa.interface import MESA
from amuse.community.evtwin.interface import EVtwin
from amuse.community.gadget2.interface import Gadget2
from amuse.community.fi.interface import Fi
from amuse.ext.star_to_sph import *
from amuse.ext.sph_to_star import convert_SPH_to_stellar_model
from amuse.units import units
from amuse.units import generic_unit_system
from amuse.units import nbody_system
from amuse.units import constants
from amuse.units.generic_unit_converter import ConvertBetweenGenericAndSiUnits
from amuse.datamodel import Particles
from amuse.datamodel import Particle
from amuse.datamodel import ParticlesSuperset
from amuse.datamodel import Grid


class TestStellarModel2SPH(TestWithMPI):

    class StarParticleWithStructure(Particle):

        def __init__(self, number_of_species=3, **keyword_arguments):
            Particle.__init__(self, **keyword_arguments)
            self.particles_set._private.number_of_species = number_of_species
            self.mass = 4.0/3.0 * numpy.pi * (9.0 / 8.0) | units.MSun
            self.radius = 1.0 | units.RSun

        def get_number_of_zones(self):
            return 4

        def get_number_of_species(self):
            return self.particles_set._private.number_of_species

        def get_names_of_species(self, number_of_species=None):
            return (['h1', 'he3', 'he4', 'c12'])[:int(self.particles_set._private.number_of_species)]

        def get_masses_of_species(self, number_of_species=None):
            return ([1.0078250, 3.0160293, 4.0026032, 12.0] | units.amu)[:int(self.particles_set._private.number_of_species)]

        def get_mass_profile(self, number_of_zones=None):
            return numpy.asarray([2.0, 14.0, 112.0, 448.0]) / sum([2.0, 14.0, 112.0, 448.0])

        def get_density_profile(self, number_of_zones=None):
            return [2.0, 2.0, 2.0, 1.0] | units.MSun/units.RSun**3

        def get_radius_profile(self, number_of_zones=None):
            return ([1.0, 2.0, 4.0, 8.0] | units.RSun) / 8.0

        def get_temperature_profile(self, number_of_zones=None):
            return [1e7, 1e6, 1e5, 1e4] | units.K

        def get_luminosity_profile(self, number_of_zones=None):
            return [1.0, 1.0, 1.0, 1.0] | units.LSun

        def get_mu_profile(self, number_of_zones=None):
            return [0.8, 0.6, 0.6, 1.3] | units.amu

        def get_chemical_abundance_profiles(self, number_of_zones=None, number_of_species=None):
            return numpy.asarray([[0.0, 0.7, 0.7, 0.7], [0.05, 0.01, 0.01, 0.01], [0.95, 0.29, 0.29, 0.29],
                [0.0, 0.0, 0.0, 0.0]])[:int(self.particles_set._private.number_of_species)]

    def test1(self):
        star = self.StarParticleWithStructure()
        number_of_zones = star.get_number_of_zones()
        delta_mass = star.mass * star.get_mass_profile()
        outer_radius = star.get_radius_profile()
        inner_radius = [0.0] | units.RSun
        inner_radius.extend(outer_radius[:-1])
        delta_radius_cubed = (outer_radius**3 - inner_radius**3)
        self.assertAlmostEqual(star.get_density_profile() / (delta_mass/(4./3.*numpy.pi*delta_radius_cubed)),
                                [1]*number_of_zones)

    def test2(self):
        star = self.StarParticleWithStructure()
        number_of_sph_particles = 100  # only few particles for test speed-up
        converter = StellarModel2SPH(star, number_of_sph_particles, seed=12345)
        converter.retrieve_stellar_structure()
        self.assertAlmostEqual(converter.specific_internal_energy_profile,
            [155896.35894, 20786.18119, 2078.61812, 95.93622] | (units.km/units.s)**2, places=1)

    def test3(self):
        print("Test interpolate_hydro_quantities")
        star = self.StarParticleWithStructure()
        number_of_sph_particles = 100  # only few particles for test speed-up
        converter = StellarModel2SPH(star, number_of_sph_particles, seed=12345)
        converter.retrieve_stellar_structure()
        outer_radii = star.get_radius_profile()
        inner_radii = [0.0] | units.RSun
        inner_radii.extend(outer_radii[:-1])
        self.assertEqual(outer_radii, [0.125, 0.25, 0.5, 1.0] | units.RSun)
        self.assertEqual(inner_radii, [0.0, 0.125, 0.25, 0.5] | units.RSun)
        radial_positions = (outer_radii + inner_radii) / 2
        int_specific_internal_energy, int_composition, int_mu = converter.interpolate_internal_energy(radial_positions)
        self.assertEqual(converter.specific_internal_energy_profile, int_specific_internal_energy)

    def test4(self):
        print("Test convert_stellar_model_to_SPH")
        star = self.StarParticleWithStructure()
        number_of_sph_particles = 100  # only few particles for test speed-up
        sph_particles = convert_stellar_model_to_SPH(
            star,
            number_of_sph_particles,
            seed=12345
        ).gas_particles
        self.assertEqual(len(sph_particles), number_of_sph_particles)
        self.assertAlmostEqual(sph_particles.mass.sum(), star.mass)
        self.assertAlmostEqual(sph_particles.center_of_mass(), [0, 0, 0] | units.RSun, 1)
        self.assertIsOfOrder(max(sph_particles.x), star.radius)
        aa = sph_particles.composition.sum(axis=1) - numpy.asarray([1.0]*number_of_sph_particles)
        self.assertAlmostEqual(sph_particles.composition.sum(axis=1), 1.0)
        self.assertTrue(numpy.all(sph_particles.h1 <= 0.7001))
        self.assertTrue(numpy.all(sph_particles.he3 <= 0.0501))
        self.assertTrue(numpy.all(sph_particles.he4 >= 0.2899))
        self.assertTrue(numpy.all(sph_particles.h1[1:] - sph_particles.h1[:-1] >= -0.0001))
        self.assertTrue(numpy.all(sph_particles.he3[1:] - sph_particles.he3[:-1] <= 0.0001))
        self.assertTrue(numpy.all(sph_particles.he4[1:] - sph_particles.he4[:-1] <= 0.0001))

    def test5(self):
        print("Test evolving created SPH particles in Gadget")
        stellar_evolution = self.new_instance_of_an_optional_code(MESA)
        if stellar_evolution is None:
            print("MESA was not built. Skipping test.")
            return
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        stellar_evolution.initialize_code()
        stellar_evolution.particles.add_particles(stars)
        stellar_evolution.commit_particles()
        number_of_sph_particles = 200  # only few particles for test speed-up
        sph_particles = convert_stellar_model_to_SPH(
            stellar_evolution.particles[0],
            number_of_sph_particles,
            seed=12345
        ).gas_particles
        self.assertEqual(len(sph_particles), number_of_sph_particles)
        self.assertAlmostEqual(sph_particles.mass.sum(), stars.mass)
        self.assertAlmostEqual(sph_particles.center_of_mass(), [0, 0, 0] | units.RSun, 1)
        self.assertIsOfOrder(max(sph_particles.x), 1 | units.RSun)
        self.assertAlmostEqual(sph_particles.h1, 0.7, places=2)
        stellar_evolution.stop()

        time_end = 1.0 | units.s
        unit_converter = ConvertBetweenGenericAndSiUnits(1.0 | units.RSun, 1.0 | units.MSun, time_end*1000)
        hydrodynamics = Gadget2(unit_converter)
        hydrodynamics.initialize_code()
        hydrodynamics.gas_particles.add_particles(sph_particles)
        hydrodynamics.evolve_model(time_end)
        hydrodynamics.stop()

    def slowtest6(self):
        print("Compare composition profile of stellar model to SPH model")
        stellar_evolution = self.new_instance(MESA)
        if stellar_evolution is None:
            print("MESA was not built. Skipping test.")
            return
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        stellar_evolution.initialize_code()
        stellar_evolution.particles.add_particles(stars)
        stellar_evolution.commit_particles()
        stellar_evolution.evolve_model(5.0 | units.Gyr)
        number_of_sph_particles = 1000  # only few particles for test speed-up
        composition = stellar_evolution.particles[0].get_chemical_abundance_profiles()
        outer_radii = stellar_evolution.particles[0].get_radius_profile()
        outer_radii.prepend(0.0 | units.m)
        midpoints = (outer_radii[:-1] + outer_radii[1:]) / 2
        sph_particles = convert_stellar_model_to_SPH(
            stellar_evolution.particles[0],
            number_of_sph_particles,
            seed=12345
        ).gas_particles
        temperature = stellar_evolution.particles[0].get_temperature_profile()
        mu = stellar_evolution.particles[0].get_mu_profile()
        specific_internal_energy = (1.5 * constants.kB * temperature / mu).as_quantity_in(units.J/units.kg)
        stellar_evolution.stop()
        sph_midpoints = sph_particles.position.lengths()

        composition_comparison_plot(
            midpoints, composition[0],
            sph_midpoints, sph_particles.h1,
            os.path.join(get_path_to_results(), "star2sph_test_6_composition_h1.png")
        )
        composition_comparison_plot(
            midpoints, composition[2],
            sph_midpoints, sph_particles.he4,
            os.path.join(get_path_to_results(), "star2sph_test_6_composition_he4.png")
        )

        time_end = 4000.0 | units.s
        unit_converter = ConvertBetweenGenericAndSiUnits(1.0 | units.RSun, 1.0 | units.MSun, time_end)
        hydrodynamics = Gadget2(unit_converter)
        hydrodynamics.initialize_code()
        hydrodynamics.gas_particles.add_particles(sph_particles)
        hydrodynamics.evolve_model(time_end)
        sph_midpoints = hydrodynamics.gas_particles.position.lengths()
        hydrodynamics.stop()
        composition_comparison_plot(
            midpoints, composition[0],
            sph_midpoints, sph_particles.h1,
            os.path.join(get_path_to_results(), "star2sph_test_6_relaxed_composition_h1.png")
        )
        composition_comparison_plot(
            midpoints, composition[2],
            sph_midpoints, sph_particles.he4,
            os.path.join(get_path_to_results(), "star2sph_test_6_relaxed_composition_he4.png")
        )
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, sph_particles.u,
            os.path.join(get_path_to_results(), "star2sph_test_6_relaxed_composition_u.png")
        )

    def slowtest7(self):
        print("Relaxation of stellar evolution model (Gadget2)")
        stellar_evolution = self.new_instance(MESA)
        if stellar_evolution is None:
            print("MESA was not built. Skipping test.")
            return
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        stellar_evolution.initialize_code()
        stellar_evolution.particles.add_particles(stars)
        stellar_evolution.commit_particles()
        stellar_evolution.evolve_model(10.0 | units.Gyr)

        composition = stellar_evolution.particles[0].get_chemical_abundance_profiles()
        outer_radii = stellar_evolution.particles[0].get_radius_profile()
        outer_radii.prepend(0.0 | units.m)
        midpoints = (outer_radii[:-1] + outer_radii[1:]) / 2
        temperature = stellar_evolution.particles[0].get_temperature_profile()
        mu = stellar_evolution.particles[0].get_mu_profile()
        specific_internal_energy = (1.5 * constants.kB * temperature / mu).as_quantity_in(units.J/units.kg)  # units.m**2/units.s**2)

        number_of_sph_particles = 10000
        print("Creating initial conditions from a MESA stellar evolution model:")
        print(stars.mass[0], "star consisting of", number_of_sph_particles, "particles.")
        gas = convert_stellar_model_to_SPH(
            stellar_evolution.particles[0],
            number_of_sph_particles,
            seed=12345
        ).gas_particles
        stellar_evolution.stop()

        t_end = 1.0e4 | units.s
        print("Evolving to:", t_end)
        n_steps = 100

        unit_converter = ConvertBetweenGenericAndSiUnits(1.0 | units.RSun, 1.0 | units.MSun, t_end)
        hydro_code = Gadget2(unit_converter)
        hydro_code.gas_particles.add_particles(gas)

        times = [] | units.Myr
        kinetic_energies = [] | units.J
        potential_energies = [] | units.J
        thermal_energies = [] | units.J
        for time in [i*t_end/n_steps for i in range(1, n_steps+1)]:
            hydro_code.evolve_model(time)
            times.append(time)
            kinetic_energies.append(hydro_code.kinetic_energy)
            potential_energies.append(hydro_code.potential_energy)
            thermal_energies.append(hydro_code.thermal_energy)

        sph_midpoints = hydro_code.gas_particles.position.lengths()
        energy_plot(times, kinetic_energies, potential_energies, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_7_after_t1e4_gadget_energy_evolution.png"))
        thermal_energy_plot(times, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_7_after_t1e4_gadget_thermal_energy_evolution.png"))
        composition_comparison_plot(
            midpoints, composition[0],
            sph_midpoints, gas.h1,
            os.path.join(get_path_to_results(), "star2sph_test_7_after_t1e4_gadget_composition_h1.png")
        )
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, gas.u,
            os.path.join(get_path_to_results(), "star2sph_test_7_after_t1e4_gadget_original_u.png")
        )
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, hydro_code.gas_particles.u,
            os.path.join(get_path_to_results(), "star2sph_test_7_after_t1e4_gadget_new_u.png")
        )
        hydro_code.stop()
        print("All done!\n")

    def slowtest8(self):
        print("Isothermal relaxation of stellar evolution model (Fi)")
        stellar_evolution = self.new_instance(MESA)
        if stellar_evolution is None:
            print("MESA was not built. Skipping test.")
            return
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        stellar_evolution.initialize_code()
        stellar_evolution.particles.add_particles(stars)
        stellar_evolution.commit_particles()
        number_of_sph_particles = 10000
        print("Creating initial conditions from a MESA stellar evolution model:")
        print(stars.mass[0], "star consisting of", number_of_sph_particles, "particles.")
        gas = convert_stellar_model_to_SPH(
            stellar_evolution.particles[0],
            number_of_sph_particles,
            seed=12345
        ).gas_particles
        stellar_evolution.stop()

        t_end = 1000.0 | units.s
        n_steps = 100

        print("Evolving to:", t_end)

        gas.h_smooth = 0.01 | units.RSun

        unit_converter = nbody_system.nbody_to_si(1000.0 | units.s, 1.0 | units.RSun)
        hydro_code = Fi(unit_converter)
        hydro_code.parameters.timestep = t_end / n_steps
        hydro_code.parameters.isothermal_flag = True
        hydro_code.parameters.gamma = 1.0
        hydro_code.gas_particles.add_particles(gas)

        times = [] | units.Myr
        kinetic_energies = [] | units.J
        potential_energies = [] | units.J
        thermal_energies = [] | units.J
        for time in [i*t_end/n_steps for i in range(1, n_steps+1)]:
            hydro_code.evolve_model(time)
            print("Evolved model to:", time)
            times.append(time)
            kinetic_energies.append(hydro_code.kinetic_energy)
            potential_energies.append(hydro_code.potential_energy)
            thermal_energies.append(hydro_code.thermal_energy)
        hydro_code.stop()
        energy_plot(times, kinetic_energies, potential_energies, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_8_fi_star2sph.png"))
        print("All done!\n")

    def slowtest9(self):
        print("Test convert_stellar_model_to_SPH and relaxation")
        stellar_evolution = self.new_instance(MESA)
        if stellar_evolution is None:
            print("MESA was not built. Skipping test.")
            return
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        stellar_evolution.initialize_code()
        stellar_evolution.particles.add_particles(stars)
        stellar_evolution.commit_particles()
        stellar_evolution.evolve_model(10.0 | units.Gyr)
        composition = stellar_evolution.particles[0].get_chemical_abundance_profiles()
        outer_radii = stellar_evolution.particles[0].get_radius_profile()
        outer_radii.prepend(0.0 | units.m)
        midpoints = (outer_radii[:-1] + outer_radii[1:]) / 2
        temperature = stellar_evolution.particles[0].get_temperature_profile()
        mu = stellar_evolution.particles[0].get_mu_profile()
        specific_internal_energy = (1.5 * constants.kB * temperature / mu).as_quantity_in(units.J/units.kg)  # units.m**2/units.s**2)

        number_of_sph_particles = 1000  # only few particles for test speed-up
        print("Creating initial conditions from a MESA stellar evolution model:")
        print(stars.mass[0], "star consisting of", number_of_sph_particles, "particles, with relaxation turned ON.")
        sph_particles = convert_stellar_model_to_SPH(
            stellar_evolution.particles[0],
            number_of_sph_particles,
            seed=12345,
            do_relax=True
        ).gas_particles
        stellar_evolution.stop()
        sph_midpoints = sph_particles.position.lengths()

        composition_comparison_plot(
            midpoints, composition[0],
            sph_midpoints, sph_particles.h1,
            os.path.join(get_path_to_results(), "star2sph_test_9_relax_h1.png")
        )
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, sph_particles.u,
            os.path.join(get_path_to_results(), "star2sph_test_9_relax_u.png")
        )

    def slowtest10(self):
        print("Test convert_stellar_model_to_SPH with relaxation, and subsequently relax with Gadget2")
        stellar_evolution = self.new_instance(MESA)
        if stellar_evolution is None:
            print("MESA was not built. Skipping test.")
            return
        stars = Particles(1)
        stars.mass = 1.0 | units.MSun
        stellar_evolution.initialize_code()
        stellar_evolution.particles.add_particles(stars)
        stellar_evolution.commit_particles()
        stellar_evolution.evolve_model(10.0 | units.Gyr)
        composition = stellar_evolution.particles[0].get_chemical_abundance_profiles()
        outer_radii = stellar_evolution.particles[0].get_radius_profile()
        outer_radii.prepend(0.0 | units.m)
        midpoints = (outer_radii[:-1] + outer_radii[1:]) / 2
        temperature = stellar_evolution.particles[0].get_temperature_profile()
        mu = stellar_evolution.particles[0].get_mu_profile()
        specific_internal_energy = (1.5 * constants.kB * temperature / mu).as_quantity_in(units.J/units.kg)  # units.m**2/units.s**2)

        number_of_sph_particles = 10000  # only few particles for test speed-up
        print("Creating initial conditions from a MESA stellar evolution model:")
        print(stars.mass[0], "star consisting of", number_of_sph_particles, "particles, with relaxation turned ON.")
        sph_particles = convert_stellar_model_to_SPH(
            stellar_evolution.particles[0],
            number_of_sph_particles,
            seed=12345,
            do_relax=True
        ).gas_particles
        stellar_evolution.stop()
        sph_midpoints = sph_particles.position.lengths()

        composition_comparison_plot(
            midpoints, composition[0],
            sph_midpoints, sph_particles.h1,
            os.path.join(get_path_to_results(), "star2sph_test_10_before_h1_new.png")
        )
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, sph_particles.u,
            os.path.join(get_path_to_results(), "star2sph_test_10_before_u_new.png")
        )
        t_end = 1.0e4 | units.s
        print("Evolving to:", t_end)
        n_steps = 100

        unit_converter = ConvertBetweenGenericAndSiUnits(1.0 | units.RSun, 1.0 | units.MSun, t_end)
        hydro_code = Gadget2(unit_converter)
        hydro_code.gas_particles.add_particles(sph_particles)

        times = [] | units.Myr
        kinetic_energies = [] | units.J
        potential_energies = [] | units.J
        thermal_energies = [] | units.J
        for time in [i*t_end/n_steps for i in range(1, n_steps+1)]:
            hydro_code.evolve_model(time)
            times.append(time)
            kinetic_energies.append(hydro_code.kinetic_energy)
            potential_energies.append(hydro_code.potential_energy)
            thermal_energies.append(hydro_code.thermal_energy)

        sph_midpoints = hydro_code.gas_particles.position.lengths()
        energy_plot(times, kinetic_energies, potential_energies, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_10_after_t1e4_gadget_energy_evolution.png"))
        thermal_energy_plot(times, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_10_after_t1e4_gadget_thermal_energy_evolution.png"))
        composition_comparison_plot(
            midpoints, composition[0],
            sph_midpoints, sph_particles.h1,
            os.path.join(get_path_to_results(), "star2sph_test_10_after_t1e4_gadget_h1.png")
        )
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, sph_particles.u,
            os.path.join(get_path_to_results(), "star2sph_test_10_after_t1e4_gadget_original_u.png")
        )
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, hydro_code.gas_particles.u,
            os.path.join(get_path_to_results(), "star2sph_test_10_after_t1e4_gadget_new_u.png")
        )
        hydro_code.stop()
        print("All done!\n")

    def test11(self):
        print("Test convert_stellar_model_to_SPH with two stars")
        star = self.StarParticleWithStructure()
        number_of_sph_particles = 100  # only few particles for test speed-up
        some_sph_particles = convert_stellar_model_to_SPH(
            star,
            number_of_sph_particles,
            seed=12345
        ).gas_particles

        another_star = self.StarParticleWithStructure(number_of_species=4)
        more_sph_particles = convert_stellar_model_to_SPH(
            another_star,
            number_of_sph_particles,
            seed=12345
        ).gas_particles
        more_sph_particles.x += 100.0 | units.RSun

        sph_particles = ParticlesSuperset([some_sph_particles, more_sph_particles])
        string_produced_by_print = sph_particles.__str__()
        self.assertTrue("he3" in string_produced_by_print)
        self.assertFalse("c12" in string_produced_by_print)

        self.assertEqual(len(sph_particles), 2 * number_of_sph_particles)
        self.assertAlmostEqual(sph_particles.mass.sum(), 2 * star.mass)
        self.assertIsOfOrder(max(sph_particles.x), (100.0 | units.RSun) + another_star.radius)
        self.assertIsOfOrder(min(sph_particles.x), -star.radius)

        self.assertEqual(len(some_sph_particles.composition), number_of_sph_particles)
        self.assertEqual(len(some_sph_particles[0].composition), 3)
        self.assertEqual(len(more_sph_particles[0].composition), 4)
        self.assertRaises(AttributeError, getattr, sph_particles, "composition",
            expected_message="Subsets return incompatible quantities for attribute 'composition', attribute cannot be queried from the superset")

        self.assertAlmostEqual(some_sph_particles.composition.sum(axis=1), [1.0]*number_of_sph_particles)
        self.assertAlmostEqual(more_sph_particles.composition.sum(axis=1), [1.0]*number_of_sph_particles)
        self.assertTrue(numpy.all(sph_particles.h1 <= 0.7001))
        self.assertTrue(numpy.all(sph_particles.he3 <= 0.0501))
        self.assertTrue(numpy.all(sph_particles.he4 >= 0.2899))

    def slowtest12(self):
        print("Test merge two stars")
        stellar_evolution = self.new_instance(MESA)
        if stellar_evolution is None:
            print("MESA was not built. Skipping test.")
            return
        stars = Particles(2)
        stars.mass = [1.0, 1.0] | units.MSun
        stellar_evolution.initialize_code()
        stellar_evolution.particles.add_particles(stars)
        stellar_evolution.commit_particles()
        stellar_evolution.evolve_model(10.0 | units.Gyr)

        composition = stellar_evolution.particles[0].get_chemical_abundance_profiles()
        outer_radii = stellar_evolution.particles[0].get_radius_profile()
        outer_radii.prepend(0.0 | units.m)
        midpoints = (outer_radii[:-1] + outer_radii[1:]) / 2
        temperature = stellar_evolution.particles[0].get_temperature_profile()
        mu = stellar_evolution.particles[0].get_mu_profile()
        specific_internal_energy = (1.5 * constants.kB * temperature / mu).as_quantity_in(units.J/units.kg)

        number_of_sph_particles = 4000
        n_string = "n4e3"
        print("Creating initial conditions from a MESA stellar evolution model:")
        print(stars.mass[0], "star consisting of", number_of_sph_particles, "particles.")
        sph_particles_1 = convert_stellar_model_to_SPH(
            stellar_evolution.particles[0],
            number_of_sph_particles,
            seed=12345
        ).gas_particles
        print(stars.mass[1], "star consisting of", number_of_sph_particles, "particles.")
        sph_particles_2 = convert_stellar_model_to_SPH(
            stellar_evolution.particles[1],
            number_of_sph_particles,
            seed=12345
        ).gas_particles
        stellar_evolution.stop()
        initial_separation = 4.0 | units.RSun
        initial_speed = 100.0 | units.km / units.s
        sph_particles_2.x += initial_separation
        sph_particles_1.vx += initial_speed
        all_sph_particles = ParticlesSuperset([sph_particles_1, sph_particles_2])

        t_end = 4.0e4 | units.s
        t_end_string = "t4e4"
        print("Evolving to:", t_end)
        n_steps = 100

        unit_converter = ConvertBetweenGenericAndSiUnits(1.0 | units.RSun, 1.0 | units.MSun, t_end)
        hydro_code = Gadget2(unit_converter)
        hydro_code.gas_particles.add_particles(all_sph_particles)

        times = [] | units.Myr
        kinetic_energies = [] | units.J
        potential_energies = [] | units.J
        thermal_energies = [] | units.J
        for time in [i*t_end/n_steps for i in range(1, n_steps+1)]:
            hydro_code.evolve_model(time)
            times.append(time)
            kinetic_energies.append(hydro_code.kinetic_energy)
            potential_energies.append(hydro_code.potential_energy)
            thermal_energies.append(hydro_code.thermal_energy)

        energy_plot(times, kinetic_energies, potential_energies, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_12_merger_"+n_string+"_"+t_end_string+"_energy_evolution.png"))
        thermal_energy_plot(times, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_12_merger_"+n_string+"_"+t_end_string+"_thermal_energy_evolution.png"))

        channel = hydro_code.gas_particles.new_channel_to(all_sph_particles)
        channel.copy_attributes(['mass', 'rho', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'u'])
        center_of_mass = all_sph_particles.center_of_mass().as_quantity_in(units.RSun)
        center_of_mass_velocity = all_sph_particles.center_of_mass_velocity().as_quantity_in(units.km / units.s)
        print("center_of_mass:", center_of_mass)
        print("center_of_mass_velocity:", center_of_mass_velocity)
        self.assertIsOfOrder(center_of_mass[0], 0.5 * (initial_separation + t_end * initial_speed))
        self.assertIsOfOrder(center_of_mass_velocity[0], 0.5 * initial_speed)
        all_sph_particles.position -= center_of_mass
        sph_midpoints = all_sph_particles.position.lengths()

        composition_comparison_plot(
            midpoints, composition[0],
            sph_midpoints, all_sph_particles.h1,
            os.path.join(get_path_to_results(), "star2sph_test_12_merger_"+n_string+"_"+t_end_string+"_composition_h1.png")
        )
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, all_sph_particles.u,
            os.path.join(get_path_to_results(), "star2sph_test_12_merger_"+n_string+"_"+t_end_string+"_new_u.png")
        )
        hydro_code.stop()
        print("All done!\n")

    def slowtest13(self):
        print("Super giant model in SPH")
        stellar_evolution = self.new_instance(MESA)
        if stellar_evolution is None:
            print("MESA was not built. Skipping test.")
            return
        star = Particle()
        star.mass = 10.0 | units.MSun
        stellar_evolution.initialize_code()
        se_star = stellar_evolution.particles.add_particle(star)
        stellar_evolution.commit_particles()
        original_outer_radii = se_star.get_radius_profile().as_quantity_in(units.RSun)
        original_density = se_star.get_density_profile()
        try:
            while True:
                stellar_evolution.evolve_model()
        except AmuseException as ex:
            self.assertEqual(str(ex), "Error when calling 'evolve' of a 'MESA', errorcode is -14, error "
            "is 'Evolve terminated: Maximum number of backups reached.'")

        composition = se_star.get_chemical_abundance_profiles()
        density = se_star.get_density_profile()
        outer_radii = se_star.get_radius_profile()
        outer_radii.prepend(0.0 | units.m)
        midpoints = (outer_radii[:-1] + outer_radii[1:]) / 2
        temperature = se_star.get_temperature_profile()
        mu = se_star.get_mu_profile()
        specific_internal_energy = (1.5 * constants.kB * temperature / mu).as_quantity_in(units.J/units.kg)

        pyplot.figure(figsize=(5, 5))
        loglog(original_outer_radii, original_density, label="t = "+str(0 | units.Myr))
        loglog(outer_radii[1:], density, label="t = "+str(se_star.age.as_quantity_in(units.Myr)))
        xlabel('radius')
        ylabel('density')
        pyplot.legend(loc=3)
        figname = os.path.join(get_path_to_results(), "star2sph_test_13_density.png")
        pyplot.savefig(figname)
        print("\nPlot of density profile was saved to: ", figname)
        pyplot.close()

        number_of_sph_particles = 1000
        print("Creating initial conditions from a MESA stellar evolution model:")
        print(star.mass[0], "star consisting of", number_of_sph_particles, "particles.")
        gas = convert_stellar_model_to_SPH(
            se_star,
            number_of_sph_particles
        ).gas_particles
        stellar_evolution.stop()

        t_end = 1.0e3 | units.s
        print("Evolving to:", t_end)
        n_steps = 100

        unit_converter = ConvertBetweenGenericAndSiUnits(1.0 | units.RSun, 1.0 | units.MSun, t_end)
        hydro_code = Gadget2(unit_converter)
        hydro_code.gas_particles.add_particles(gas)

        times = [] | units.Myr
        kinetic_energies = [] | units.J
        potential_energies = [] | units.J
        thermal_energies = [] | units.J
        for time in [i*t_end/n_steps for i in range(1, n_steps+1)]:
            hydro_code.evolve_model(time)
            times.append(time)
            kinetic_energies.append(hydro_code.kinetic_energy)
            potential_energies.append(hydro_code.potential_energy)
            thermal_energies.append(hydro_code.thermal_energy)

        sph_midpoints = hydro_code.gas_particles.position.lengths()
        energy_plot(times, kinetic_energies, potential_energies, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_13_n1e3_after_t1e3_gadget_energy_evolution.png"))
        thermal_energy_plot(times, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_13_n1e3_after_t1e3_gadget_thermal_energy_evolution.png"))
        composition_comparison_plot(
            midpoints, composition[0],
            sph_midpoints, gas.h1,
            os.path.join(get_path_to_results(), "star2sph_test_13_n1e3_after_t1e3_gadget_composition_h1.png")
        )
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, gas.u,
            os.path.join(get_path_to_results(), "star2sph_test_13_n1e3_after_t1e3_gadget_original_u.png")
        )
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, hydro_code.gas_particles.u,
            os.path.join(get_path_to_results(), "star2sph_test_13_n1e3_after_t1e3_gadget_new_u.png")
        )
        hydro_code.stop()
        print("All done!\n")

    def slowtest14(self):
        print("SPH model with core")
        # options:
        with_core = True  # set to False to do a comparison run without a core (True)
        hydro_code = Gadget2  # Fi -or- Gadget2

        test_pickle_file = os.path.join(get_path_to_results(), "star2sph_test_14.pkl")
        if not os.path.exists(test_pickle_file):
            stellar_evolution = self.new_instance(MESA)
            star = stellar_evolution.particles.add_particle(Particle(mass=5.0 | units.MSun))
            while star.radius < (200 | units.RSun):
                star.evolve_for(10 * star.time_step)
            print(star.core_mass)
            pickle_stellar_model(star, test_pickle_file)
            stellar_evolution.stop()

        number_of_sph_particles = 10000
        print("Creating initial conditions from a MESA stellar evolution model")

        converter = StellarModel2SPH(
            None,
            number_of_sph_particles,
            seed=12345,
            pickle_file=test_pickle_file,
            with_core_particle=with_core,
            target_core_mass=0.9 | units.MSun
        )
        converter.unpickle_stellar_structure()
        original_mass = converter.mass
        stellar_model_in_SPH = converter.result

        if with_core:
            print("Created", len(stellar_model_in_SPH.gas_particles), end=' ')
            print("SPH particles and one 'core-particle':\n", stellar_model_in_SPH.core_particle)
            print(stellar_model_in_SPH.gas_particles.mass.sum())
            core_radius = stellar_model_in_SPH.core_radius
        else:
            print("Only SPH particles created.")
            core_radius = 1.0 | units.RSun
        print("Setting gravitational smoothing to:", core_radius.as_quantity_in(units.RSun))

        t_end = 1.0e4 | units.s
        n_steps = 100

        unit_converter = ConvertBetweenGenericAndSiUnits(1.0 | units.RSun, constants.G, t_end)
        hydro_code = hydro_code(unit_converter)
        try:
            hydro_code.parameters.timestep = t_end / n_steps
        except Exception as exc:
            if not "parameter is read-only" in str(exc):
                raise
        hydro_code.parameters.epsilon_squared = core_radius**2
        hydro_code.gas_particles.add_particles(stellar_model_in_SPH.gas_particles)
        if with_core:
            hydro_code.dm_particles.add_particle(stellar_model_in_SPH.core_particle)

        self.assertAlmostRelativeEqual(original_mass, hydro_code.particles.total_mass(), places=7)

        sph_midpoints = hydro_code.gas_particles.position.lengths()
        radial_comparison_plot(
            converter.midpoints_profile[1:-1], converter.specific_internal_energy_profile,
            sph_midpoints, hydro_code.gas_particles.u,
            os.path.join(get_path_to_results(), "star2sph_test_14_internal_energy.png"),
            y_label="internal energy"
        )
        radial_comparison_plot(
            converter.midpoints_profile[1:-1], converter.density_profile,
            sph_midpoints, hydro_code.gas_particles.rho,
            os.path.join(get_path_to_results(), "star2sph_test_14_density.png"),
            y_label="density"
        )
        radial_comparison_plot(
            converter.midpoints_profile[1:-1], converter.specific_internal_energy_profile * converter.density_profile**(-2.0/3.0),
            sph_midpoints, hydro_code.gas_particles.u * hydro_code.gas_particles.rho**(-2.0/3.0),
            os.path.join(get_path_to_results(), "star2sph_test_14_entropy.png"),
            y_label="entropy"
        )
        times = [] | units.s
        kinetic_energies = [] | units.J
        potential_energies = [] | units.J
        thermal_energies = [] | units.J

        print("Evolving to:", t_end)
        for time in [i*t_end/n_steps for i in range(1, n_steps+1)]:
            hydro_code.evolve_model(time)
            times.append(time)
            kinetic_energies.append(hydro_code.kinetic_energy)
            potential_energies.append(hydro_code.potential_energy)
            thermal_energies.append(hydro_code.thermal_energy)

        sph_midpoints = hydro_code.gas_particles.position.lengths()
        energy_plot(times, kinetic_energies, potential_energies, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_14_after_t1e2_gadget_energy_evolution.png"))
        thermal_energy_plot(times, thermal_energies,
            os.path.join(get_path_to_results(), "star2sph_test_14_after_t1e2_gadget_thermal_energy_evolution.png"))
        radial_comparison_plot(
            converter.midpoints_profile[1:-1], converter.specific_internal_energy_profile,
            sph_midpoints, hydro_code.gas_particles.u,
            os.path.join(get_path_to_results(), "star2sph_test_14_after_t1e2_gadget_internal_energy.png"),
            y_label="internal energy"
        )
        hydro_code.stop()
        print("All done!\n")

    def test15(self):
        print("Test pickling of stellar structure")
        star = self.StarParticleWithStructure()
        test_pickle_file = os.path.join(get_path_to_results(), "test_star_structure.pkl")
        if os.path.exists(test_pickle_file):
            os.remove(test_pickle_file)
        pickle_stellar_model(star, test_pickle_file)
        converter = StellarModel2SPH(None, 100, seed=12345,
            pickle_file=test_pickle_file)
        converter.unpickle_stellar_structure()
        self.assertEqual(converter.mass, numpy.pi * 1.5 | units.MSun)
        self.assertEqual(converter.radius, 1.0 | units.RSun)
        self.assertEqual(converter.number_of_zones, 4)
        self.assertEqual(converter.number_of_species, 3)
        self.assertEqual(converter.species_names, ['h1', 'he3', 'he4'])
        self.assertEqual(converter.density_profile, [2.0, 2.0, 2.0, 1.0] | units.MSun/units.RSun**3)
        self.assertEqual(converter.radius_profile, [1.0/8.0, 1.0/4.0, 1.0/2.0, 1.0] | units.RSun)
# ~        self.assertEqual(converter.temperature_profile, [1e7, 1e6, 1e5, 1e4] | units.K)
        self.assertEqual(converter.mu_profile, [0.8, 0.6, 0.6, 1.3] | units.amu)
        self.assertEqual(converter.composition_profile, [[0.0, 0.7, 0.7, 0.7],
            [0.05, 0.01, 0.01, 0.01], [0.95, 0.29, 0.29, 0.29]])
        self.assertAlmostEqual(converter.specific_internal_energy_profile,
            [155896.35894, 20786.18119, 2078.61812, 95.93622] | (units.km/units.s)**2, places=1)

        self.assertRaises(AmuseWarning, pickle_stellar_model, star, test_pickle_file, expected_message="Incorrect file name '{0}'; directory must exist and file may not exist".format(test_pickle_file))
        bogus_pickle_file = os.path.join(get_path_to_results(), "bogus.pkl")
        converter = StellarModel2SPH(None, 100, seed=12345,
            pickle_file=bogus_pickle_file)
        self.assertRaises(AmuseException, converter.unpickle_stellar_structure, expected_message="Input pickle file '{0}' does not exist".format(bogus_pickle_file))

    def test16(self):
        print("Test convert_stellar_model_to_SPH with pickled stellar structure")
        star = self.StarParticleWithStructure()
        test_pickle_file = os.path.join(get_path_to_results(), "test_star_structure.pkl")
        if os.path.exists(test_pickle_file):
            os.remove(test_pickle_file)
        pickle_stellar_model(star, test_pickle_file)

        number_of_sph_particles = 100  # only few particles for test speed-up
        sph_particles = convert_stellar_model_to_SPH(
            None,
            number_of_sph_particles,
            seed=12345,
            pickle_file=test_pickle_file
        ).gas_particles
        self.assertEqual(len(sph_particles), number_of_sph_particles)
        self.assertAlmostEqual(sph_particles.mass.sum(), star.mass)
        self.assertAlmostEqual(sph_particles.center_of_mass(), [0, 0, 0] | units.RSun, 1)
        self.assertIsOfOrder(max(sph_particles.x), star.radius)
        self.assertAlmostEqual(sph_particles.composition.sum(axis=1), [1.0]*number_of_sph_particles)
        self.assertTrue(numpy.all(sph_particles.h1 <= 0.7001))
        self.assertTrue(numpy.all(sph_particles.he3 <= 0.0501))
        self.assertTrue(numpy.all(sph_particles.he4 >= 0.2899))
        self.assertTrue(numpy.all(sph_particles.h1[1:] - sph_particles.h1[:-1] >= -0.0001))
        self.assertTrue(numpy.all(sph_particles.he3[1:] - sph_particles.he3[:-1] <= 0.0001))
        self.assertTrue(numpy.all(sph_particles.he4[1:] - sph_particles.he4[:-1] <= 0.0001))

    def slowtest17(self):
        print("SPH red super giant model with core")
        # options:
        with_core = True  # set to False to do a comparison run without a core (True)
        use_hydro_code = Gadget2  # Fi -or- Gadget2
        use_stellar_evolution_code = MESA  # to be implemented as option...
        number_of_sph_particles = 3000
        t_end = 3.0e6 | units.s

        # Convert some of the parameters to string, for use in output file names:
        hydro_code_string = "_" + str(use_hydro_code.__name__)
        n_string = "_n" + ("%1.0e" % (number_of_sph_particles)).replace("+0", "").replace("+", "")
        t_end_string = "_t" + ("%1.0e" % (t_end.value_in(units.s))).replace("+0", "").replace("+", "")
        base_plotfile_string = "star2sph_test_17" + n_string + hydro_code_string + t_end_string

        stellar_evolution = self.new_instance(MESA, redirection="none")
        if stellar_evolution is None:
            print("MESA was not built. Skipping test.")
            return
        stars = Particles(1)
        stars.mass = 10.0 | units.MSun
        stellar_evolution.initialize_code()
        stellar_evolution.particles.add_particles(stars)
        stellar_evolution.commit_particles()

        star_age = [] | units.yr
        star_radius = [] | units.RSun
        try:
            while True:
                stellar_evolution.evolve_model()
                star_age.append(stellar_evolution.particles[0].age)
                star_radius.append(stellar_evolution.particles[0].radius)
        except AmuseException as ex:
            self.assertEqual(str(ex), "Error when calling 'evolve' of a 'MESA', errorcode is -14, error "
            "is 'Evolve terminated: Maximum number of backups reached.'")

        radius_evolution_plot(star_age, star_radius,
            os.path.join(get_path_to_results(), base_plotfile_string + "_radius_evolution.png"))
        self.assertIsOfOrder(stellar_evolution.particles[0].age,  # MS lifetime:
            (1.0e10 | units.yr) * (stars.mass.value_in(units.MSun)) ** -2.5)

        outer_radii = stellar_evolution.particles[0].get_radius_profile()
        outer_radii.prepend(0.0 | units.m)
        midpoints = (outer_radii[:-1] + outer_radii[1:]) / 2
        temperature = stellar_evolution.particles[0].get_temperature_profile()
        mu = stellar_evolution.particles[0].get_mu_profile()
        specific_internal_energy = (1.5 * constants.kB * temperature / mu).as_quantity_in(units.J/units.kg)  # units.m**2/units.s**2)

        print("Creating initial conditions from a MESA stellar evolution model:")
        print(stars.mass[0], "star consisting of", number_of_sph_particles, "particles.")

        stellar_model_in_SPH = convert_stellar_model_to_SPH(
            stellar_evolution.particles[0],
            number_of_sph_particles,
            seed=12345,
            base_grid_options=dict(type="glass", target_rms=0.04),
            with_core_particle=with_core
        )
        if len(stellar_model_in_SPH.core_particle):
            print("Created", len(stellar_model_in_SPH.gas_particles), end=' ')
            print("SPH particles and one 'core-particle':\n", stellar_model_in_SPH.core_particle)
            core_radius = stellar_model_in_SPH.core_radius
        else:
            print("Only SPH particles created.")
            core_radius = 1.0 | units.RSun
        print("Setting gravitational smoothing to:", core_radius)

        t_dyn = (stellar_evolution.particles[0].radius**3 / (2*constants.G*stars.mass[0])).sqrt()
        print("Dynamical timescale:", t_dyn.as_quantity_in(units.yr))
        stellar_evolution.stop()

        print("Evolving to:", t_end, "("+str((t_end/t_dyn)), "dynamical timescales)")
        n_steps = 100

        unit_converter = ConvertBetweenGenericAndSiUnits(1.0 | units.RSun, constants.G, t_end)
        hydro_code = use_hydro_code(unit_converter, redirection="none")

        try:
            hydro_code.parameters.timestep = t_end / n_steps
        except Exception as exc:
            if not "parameter is read-only" in str(exc):
                raise

        hydro_code.parameters.epsilon_squared = core_radius**2
        hydro_code.gas_particles.add_particles(stellar_model_in_SPH.gas_particles)
        if len(stellar_model_in_SPH.core_particle):
            hydro_code.dm_particles.add_particles(stellar_model_in_SPH.core_particle)

        self.assertAlmostRelativeEqual(stars.mass, hydro_code.particles.total_mass(), places=7)

        times = [] | units.s
        kinetic_energies = [] | units.J
        potential_energies = [] | units.J
        thermal_energies = [] | units.J
        for time in [i*t_end/n_steps for i in range(1, n_steps+1)]:
            hydro_code.evolve_model(time)
            times.append(time)
            kinetic_energies.append(hydro_code.kinetic_energy)
            potential_energies.append(hydro_code.potential_energy)
            thermal_energies.append(hydro_code.thermal_energy)

        sph_midpoints = hydro_code.gas_particles.position.lengths()
        energy_plot(times, kinetic_energies, potential_energies, thermal_energies,
            os.path.join(get_path_to_results(), base_plotfile_string + "_energy_evolution.png"))
        thermal_energy_plot(times, thermal_energies,
            os.path.join(get_path_to_results(), base_plotfile_string + "_thermal_energy_evolution.png"))
        internal_energy_comparison_plot(
            midpoints, specific_internal_energy,
            sph_midpoints, hydro_code.gas_particles.u,
            os.path.join(get_path_to_results(), base_plotfile_string + "_internal_energy.png")
        )
        hydro_code.stop()
        print("All done!\n")

    def slowtest18(self):
        print("SPH red super giant model with core (fixed core mass)")
        number_of_sph_particles = 300

        stellar_evolution = self.new_instance(MESA, redirection="none")
        if stellar_evolution is None:
            print("MESA was not built. Skipping test.")
            return
        stars = Particles(1)
        stars.mass = 50.0 | units.MSun
        stellar_evolution.initialize_code()
        stellar_evolution.particles.add_particles(stars)
        stellar_evolution.commit_particles()
        stellar_evolution.evolve_model(3.927 | units.Myr)

        expected_core_radii = [0.37648393, 0.58139942, 3.26189210, 31.89893263] | units.RSun
        for i, tgt_core_mass in enumerate([1.0, 5.0, 25.0, 49.0] | units.MSun):
            stellar_model_in_SPH = convert_stellar_model_to_SPH(
                stellar_evolution.particles[0],
                number_of_sph_particles,
                seed=12345,
                with_core_particle=True,
                target_core_mass=tgt_core_mass
            )
            self.assertAlmostRelativeEqual(stellar_model_in_SPH.core_particle[0].mass, tgt_core_mass, 1)
            self.assertAlmostEqual(stellar_model_in_SPH.core_radius, expected_core_radii[i])
        stellar_evolution.stop()

    def test19(self):
        print("Test convert_stellar_model_to_SPH with do_store_composition")
        star = self.StarParticleWithStructure(number_of_species=4)
        number_of_sph_particles = 100  # only few particles for test speed-up

        sph_particles = convert_stellar_model_to_SPH(
            star,
            number_of_sph_particles,
            seed=12345,
            do_store_composition=False
        ).gas_particles
        self.assertEqual(len(sph_particles), number_of_sph_particles)
        self.assertTrue(set(sph_particles.get_attribute_names_defined_in_store()) ==
            set(["mass", "x", "y", "z", "vx", "vy", "vz", "u", "h_smooth"]))
        self.assertTrue(set(sph_particles.get_attribute_names_defined_in_store()).isdisjoint(
            set(["h1", "he3", "he4", "c12"])))

        sph_particles = convert_stellar_model_to_SPH(
            star,
            number_of_sph_particles,
            seed=12345,
        ).gas_particles
        self.assertEqual(len(sph_particles), number_of_sph_particles)
        self.assertTrue(set(sph_particles.get_attribute_names_defined_in_store()) >
            set(["mass", "x", "y", "z", "vx", "vy", "vz", "u", "h_smooth"]))
        self.assertTrue(set(sph_particles.get_attribute_names_defined_in_store()) >
            set(["h1", "he3", "he4", "c12"]))

    def test20(self):
        print("Test convert_stellar_model_to_SPH with with_core_particle")
        star = self.StarParticleWithStructure(number_of_species=4)
        number_of_sph_particles = 100  # only few particles for test speed-up

        self.assertRaises(AmuseException, convert_stellar_model_to_SPH,
            star,
            number_of_sph_particles,
            seed=12345,
            with_core_particle=True,
            expected_message="Requested model has with_core_particle=True, but no target_core_mass specified."
        )

        star.core_mass = 0.5 | units.MSun
        self.assertRaises(AmuseException, convert_stellar_model_to_SPH,
            star,
            number_of_sph_particles,
            seed=12345,
            with_core_particle=True,
            expected_message="Requested target_core_mass of 0.5 MSun is out of range."
        )

        stellar_evolution = EVtwin()
        star = stellar_evolution.particles.add_particle(Particle(mass=5 | units.MSun))
        model = convert_stellar_model_to_SPH(
            star,
            number_of_sph_particles,
            seed=12345,
            with_core_particle=True,
            target_core_mass=0.5 | units.MSun
        )
        stellar_evolution.stop()
        core = model.core_particle
        sph_particles = model.gas_particles

        self.assertEqual(len(sph_particles), number_of_sph_particles)
        self.assertTrue(isinstance(core, Particle))

        self.assertEqual(core.radius, model.core_radius)
        self.assertEqual(core.position, [0, 0, 0] | units.m)
        self.assertAlmostEqual(core.mass + sph_particles.total_mass(), 5.0 | units.MSun)
        self.assertIsOfOrder(core.mass, 0.5 | units.MSun)

        self.assertEqual(set(sph_particles.get_attribute_names_defined_in_store()),
            set(["mass", "x", "y", "z", "vx", "vy", "vz", "u", "h_smooth",
                "h1", "he4", "c12", "n14", "o16", "ne20", "mg24", "si28", "fe56", "mu"]))

    def test21(self):
        print("Test convert_stellar_model_to_SPH and back")
        star = self.StarParticleWithStructure()
        number_of_sph_particles = 50000  # only few particles for test speed-up
        sph_particles = convert_stellar_model_to_SPH(
            star,
            number_of_sph_particles,
            base_grid_options=dict(type="fcc")
        ).gas_particles
        self.assertEqual(len(sph_particles), number_of_sph_particles)
        self.assertAlmostEqual(sph_particles.mass.sum(), star.mass)
        unit_converter = ConvertBetweenGenericAndSiUnits(1.0 | units.RSun, 1.0 | units.MSun, 1.0 | units.s)
        hydrodynamics = Gadget2(unit_converter)
        in_hydro = hydrodynamics.gas_particles.add_particles(sph_particles)
        sph_particles.pressure = in_hydro.pressure
        sph_particles.density = in_hydro.density
        model = convert_SPH_to_stellar_model(
            sph_particles, particles_per_zone=100
        #    particles_per_zone=number_of_sph_particles/4
        )
        self.assertAlmostEqual(model.dmass.sum(), star.mass)
        self.assertAlmostEqual(model.radius[-1], star.radius, 1)

#        for a,b in model.iteritems():
#            print a,b
        print(model.dmass[0].as_quantity_in(units.MSun))
        print(model.rho[0:10].as_quantity_in(units.MSun/units.RSun**3))
        print(model.rho[-10:].as_quantity_in(units.MSun/units.RSun**3))
        print(model.radius[0:10].as_quantity_in(units.RSun))
        print(model.X_H[0:10])
        print(model.X_He[0:10])


def composition_comparison_plot(radii_SE, comp_SE, radii_SPH, comp_SPH, figname):
    if not HAS_MATPLOTLIB:
        return
    pyplot.figure(figsize=(7, 5))
    plot(radii_SE.as_quantity_in(units.RSun), comp_SE,
        label='stellar evolution model')
    plot(radii_SPH, comp_SPH, 'go', label='SPH model')
    xlabel('radius')
    ylabel('mass fraction')
    pyplot.legend()
    pyplot.savefig(figname)
    print("\nPlot of composition profiles was saved to: ", figname)
    pyplot.close()


def radial_comparison_plot(radii_SE, u_SE, radii_SPH, u_SPH, figname, y_label="quantity"):
    if not HAS_MATPLOTLIB:
        return
    pyplot.figure(figsize=(7, 5))
    semilogy(radii_SE.as_quantity_in(units.RSun), u_SE,
        label='stellar evolution model')
    semilogy(radii_SPH, u_SPH, 'go', label='SPH model')
    xlabel('radius')
    ylabel(y_label)
    pyplot.legend(loc="lower center")
    pyplot.savefig(figname)
    print("\nPlot of", y_label, "profiles was saved to: ", figname)
    pyplot.close()


def internal_energy_comparison_plot(radii_SE, u_SE, radii_SPH, u_SPH, figname):
    radial_comparison_plot(radii_SE, u_SE, radii_SPH, u_SPH, figname, y_label="internal energy")


def energy_plot(time, E_kin, E_pot, E_therm, figname):
    if not HAS_MATPLOTLIB:
        return
    pyplot.figure(figsize=(5, 5))
    plot(time, E_kin.as_quantity_in(units.erg), label='E_kin')
    plot(time, E_pot, label='E_pot')
    plot(time, E_therm, label='E_therm')
    plot(time, E_kin+E_pot+E_therm, label='E_total')
    xlabel('Time')
    ylabel('Energy')
    pyplot.legend(loc=3)
    pyplot.savefig(figname)
    print("\nPlot of energy evolution was saved to: ", figname)
    pyplot.close()


def thermal_energy_plot(time, E_therm, figname):
    if not HAS_MATPLOTLIB:
        return
    pyplot.figure(figsize=(5, 5))
    plot(time, E_therm.as_quantity_in(units.erg), label='E_therm')
    xlabel('Time')
    ylabel('Energy')
    pyplot.legend(loc=3)
    pyplot.savefig(figname)
    print("\nPlot of thermal energy evolution was saved to: ", figname)
    pyplot.close()


def radius_evolution_plot(star_age, star_radius, figname):
    if not HAS_MATPLOTLIB:
        return
    pyplot.figure(figsize=(5, 5))
    plot(star_age, star_radius)
    xlabel('Time')
    ylabel('Radius')
    pyplot.savefig(figname)
    print("\nPlot of radius evolution was saved to: ", figname)
    pyplot.close()



from amuse.support.interface import InCodeComponentImplementation

from amuse.test.amusetest import TestWithMPI
from amuse.test import compile_tools
from amuse.support import exceptions

import subprocess
import os
import time
import shutil
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse import config
from amuse.rfi.tools import create_java
from amuse.rfi import channel
from amuse.rfi.core import *
codestring = """
public class Code implements CodeInterface {

  public Code(String codeDir, String amuseRootDir) {
  }

  public void end() {
        //IGNORE
  }
  
  // parameter "int_in" is an input parameter
  // parameter "int_out" is an output parameter
  // parameter "len" is a length parameter
  public int echo_array_with_result(int[] int_in, int[] int_out, int len) {
	for(int i = 0; i < len; i++) {
	    int_out[i] = int_in[i];
	}
	return -1;
  }
  
  
  // parameter "string_in" is an input parameter
  public int print_error_string(String[] string_in) {
	for(int i = 0; i < string_in.length; i++) {
    	    System.err.println(string_in[i]);
	}
	return 0;
  }
  
  
  // parameter "string_inout1" is an inout parameter
  // parameter "string_inout2" is an inout parameter
  public int echo_strings(String[] string_inout1, String[] string_inout2) {
	for(int i = 0; i < string_inout1.length; i++) {
	    String tmp = string_inout1[i];
	    string_inout1[i] = string_inout2[i];
	    string_inout2[i] = tmp;
	}
	return 0;
  }
  
  
  // parameter "double_in" is an input parameter
  // parameter "double_out" is an output parameter
  public int echo_double(double[] double_in, double[] double_out) {
	for(int i = 0; i < double_in.length; i++) {
	    double_out[i] = double_in[i];
	}
	return 0;
  }
  
  
  // parameter "float_in" is an input parameter
  // parameter "float_out" is an output parameter
  public int echo_float(float[] float_in, float[] float_out) {
	for(int i = 0; i < float_in.length; i++) {
	    float_out[i] = float_in[i];
	}
	return 0;
  }
  
  
  // parameter "int_in" is an input parameter
  // parameter "int_out" is an output parameter
  public int echo_int(int int_in, int[] int_out) {
	int_out[0] = int_in;
	if (int_in < 0) {
	    return -1;
	} else {
   	    return 0;
	}
  }

  // parameter "int_in" is an input parameter
  // parameter "int_out" is an output parameter
  public int echo_int_array(int[] int_in, int[] int_out) {
	for(int i = 0; i < int_in.length; i++) {
	    int_out[i] = int_in[i];
	}
	return 0;
  }
   
  
  // parameter "in_out" is an inout parameter
  // parameter "len" is a length parameter
  public int echo_inout_array_with_result(int[] in_out, int len) {
      for(int i = 0; i < len; i++) {
          in_out[i] = in_out[i] + 10;
      }
      return 11;
  }
  
  
  // parameter "int_in" is an input parameter
  // parameter "int_out" is an output parameter
  // parameter "len" is a length parameter
  public void echo_array(int[] int_in, int[] int_out, int len) {
	for(int i = 0; i < len; i++) {
	    int_out[i] = int_in[i];
	}
  }
  
  
  // parameter "in" is an input parameter
  // parameter "out" is an output parameter
  public int echo_long_long_int(long[] in, long[] out) {
	for(int i = 0; i < in.length; i++) {
 	    out[i] = in[i];
	}
	if (in[0] < 0) {
	    return -1;
	} else {
   	    return 0;
	}
 }
  
  
  // parameter "string_in" is an input parameter
  // parameter "string_out" is an output parameter
  public int echo_string(String[] string_in, String[] string_out) {
	for(int i = 0; i < string_in.length; i++) {
	    string_out[i] = string_in[i];
	}
	return 0;
  }
  
  
  // parameter "input" is an input parameter
  // parameter "output" is an output parameter
  public int echo_logical(boolean[] input, boolean[] output) {
	for(int i = 0; i < input.length; i++) {
	    output[i] = input[i];
	}
	return 0;
  }
  
  
  // parameter "string_in" is an input parameter
  public int print_string(String[] string_in) {
	for(int i = 0; i < string_in.length; i++) {
    	    System.out.println(string_in[i]);
	}
	return 0;
  }
  
  
  // parameter "int_in1" is an input parameter
  // parameter "int_in2" is an input parameter
  // parameter "int_out1" is an output parameter
  // parameter "int_out2" is an output parameter
  // parameter "len" is a length parameter
  public int echo_2_int(int[] int_in1, int[] int_in2, int[] int_out1, int[] int_out2, int len) {
	for(int i = 0; i < len; i++) {
            int_out1[i] = int_in1[i];
            int_out2[i] = int_in2[i];
    	}

	return len;
  }
  
}
"""


class ForTestingInterface(CodeInterface):

    classpath = ['worker.jar']
    cwd = ''

    def __init__(self, exefile, **options):
        CodeInterface.__init__(self, exefile, **options)

    # override default channel_type. We don't support the mpi channel...
    @option(choices=['mpi', 'remote', 'distributed', 'sockets'], sections=("channel",))
    def channel_type(self):
        return 'sockets'

    @legacy_function
    def echo_int():
        function = LegacyFunctionSpecification()
        function.addParameter('int_in', dtype='int32', direction=function.IN)
        function.addParameter('int_out', dtype='int32', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def echo_int_array():
        function = LegacyFunctionSpecification()
        function.addParameter('int_in', dtype='int32', direction=function.IN)
        function.addParameter('int_out', dtype='int32', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def echo_long_long_int():
        function = LegacyFunctionSpecification()
        function.addParameter('in', dtype='int64', direction=function.IN)
        function.addParameter('out', dtype='int64', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def echo_double():
        function = LegacyFunctionSpecification()
        function.addParameter('double_in', dtype='float64', direction=function.IN)
        function.addParameter('double_out', dtype='float64', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def echo_float():
        function = LegacyFunctionSpecification()
        function.addParameter('float_in', dtype='float32', direction=function.IN)
        function.addParameter('float_out', dtype='float32', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def echo_string():
        function = LegacyFunctionSpecification()
        function.addParameter('string_in', dtype='string', direction=function.IN)
        function.addParameter('string_out', dtype='string', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def echo_strings():
        function = LegacyFunctionSpecification()
        function.addParameter('string_inout1', dtype='string', direction=function.INOUT)
        function.addParameter('string_inout2', dtype='string', direction=function.INOUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def echo_array():
        function = LegacyFunctionSpecification()
        function.addParameter('int_in', dtype='int32', direction=function.IN)
        function.addParameter('int_out', dtype='int32', direction=function.OUT)
        function.addParameter('len', dtype='int32', direction=function.LENGTH)
        function.result_type = None
        function.must_handle_array = True
        return function

    @legacy_function
    def echo_array_with_result():
        function = LegacyFunctionSpecification()
        function.addParameter('int_in', dtype='int32', direction=function.IN)
        function.addParameter('int_out', dtype='int32', direction=function.OUT)
        function.addParameter('len', dtype='int32', direction=function.LENGTH)
        function.result_type = 'int32'
        function.must_handle_array = True
        return function

    # @legacy_function
    def return_string():
        function = LegacyFunctionSpecification()
        function.addParameter('string_in', dtype='string', direction=function.IN)
        function.result_type = 'string'
        function.can_handle_array = True
        return function

    @legacy_function
    def echo_2_int():
        function = LegacyFunctionSpecification()
        function.addParameter('int_in1', dtype='int32', direction=function.IN)
        function.addParameter('int_in2', dtype='int32', direction=function.IN, default=1)
        function.addParameter('int_out1', dtype='int32', direction=function.OUT)
        function.addParameter('int_out2', dtype='int32', direction=function.OUT)
        function.addParameter('len', dtype='int32', direction=function.LENGTH)
        function.result_type = 'int32'
        function.must_handle_array = True
        return function

    @legacy_function
    def echo_inout_array_with_result():
        function = LegacyFunctionSpecification()
        function.addParameter('in_out', dtype='int32', direction=function.INOUT)
        function.addParameter('len', dtype='int32', direction=function.LENGTH)
        function.result_type = 'int32'
        function.must_handle_array = True
        return function

    @legacy_function
    def echo_logical():
        function = LegacyFunctionSpecification()
        function.addParameter('input', dtype='bool', direction=function.IN)
        function.addParameter('output', dtype='bool', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def print_string():
        function = LegacyFunctionSpecification()
        function.addParameter('string_in', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def print_error_string():
        function = LegacyFunctionSpecification()
        function.addParameter('string_in', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function


class ForTesting(InCodeComponentImplementation):

    def __init__(self, exefile, **options):
        InCodeComponentImplementation.__init__(self, ForTestingInterface(exefile, **options), **options)

    def define_methods(self, object):
        object.add_method(
            'echo_int',
            (units.m,),
            (
     		units.m,
                object.ERROR_CODE,
            )
        )


class TestInterface(TestWithMPI):

    @classmethod
    def check_not_in_mpiexec(cls):
        """
        The tests will fork another process, if the test run 
        is itself an mpi process, the tests may fail.

        For the hydra process manager the tests will fail.
        So skip the tests if we detect hydra
        """
        if 'HYDI_CONTROL_FD' in os.environ:
            return  # can run in modern mpiexec.hydra
        if 'HYDRA_CONTROL_FD' in os.environ or 'PMI_FD' in os.environ:
            cls.skip('cannot run the socket tests under mpi process manager')

    @classmethod
    def check_has_java(cls):
        """
        The tests will fork another process, if the test run 
        is itself an mpi process, the tests may fail.

        For the hydra process manager the tests will fail.
        So skip the tests if we detect hydra
        """

        if not config.java.is_enabled:
            cls.skip("java not enabled")

        javac = config.java.javac
        if not os.path.exists(javac):
            cls.skip("java compiler not available")

    @classmethod
    def setup_class(cls):
        cls.check_not_in_mpiexec()
        cls.check_has_java()
        cls.check_can_compile_modules()
        try:
            cls.exefile = compile_tools.build_java_worker(codestring, cls.get_path_to_results(), ForTestingInterface)
        except Exception as ex:
            print(ex)
            raise

    def test1(self):
        instance = ForTestingInterface(self.exefile)
        int_out, error = instance.echo_int(10)
        instance.stop()
        self.assertEqual(int_out, 10)
        self.assertEqual(error, 0)

    def test2(self):
        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_double(4.0)
        instance.stop()
        self.assertEqual(out, 4.0)
        self.assertEqual(error, 0)

    def test3(self):
        instance = ForTestingInterface(self.exefile)
        input = [1, 2, 3, 4]
        output, errors = instance.echo_int_array(input)
        instance.stop()
        self.assertEqual(len(errors), 4)
        for actual, expected in zip(output, input):
            self.assertEqual(actual, expected)

    def test4(self):
        instance = ForTestingInterface(self.exefile)
        input = [1.0, 2.1, 3.3, 4.2]
        output, errors = instance.echo_double(input)
        instance.stop()
        self.assertEqual(len(errors), 4)
        for actual, expected in zip(output, input):
            self.assertEqual(actual, expected)

    def test5(self):
        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_float(4.0)
        instance.stop()
        self.assertEqual(out, 4.0)
        self.assertEqual(error, 0)

    def test6(self):
        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_string("abc")
        instance.stop()
        self.assertEqual(error, 0)
        self.assertEqual(out, "abc")

    def test7(self):
        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_string(["abc", "def"])
        instance.stop()

        self.assertEqual(error[0], 0)
        self.assertEqual(error[1], 0)
        self.assertEqual(out[0], "abc")
        self.assertEqual(out[1], "def")

    def test8(self):
        instance = ForTestingInterface(self.exefile)
        out1, out2, error = instance.echo_strings("abc", "def")
        instance.stop()

        self.assertEqual(error, 0)
        self.assertEqual(out1, "def")
        self.assertEqual(out2, "abc")

    def test9(self):
        instance = ForTestingInterface(self.exefile)
        str1_out, str2_out, error = instance.echo_strings(["abc", "def"], ["ghi", "jkl"])
        instance.stop()

        self.assertEqual(error[0], 0)
        self.assertEqual(error[1], 0)
        self.assertEqual(str1_out[0], "ghi")
        self.assertEqual(str1_out[1], "jkl")
        self.assertEqual(str2_out[0], "abc")
        self.assertEqual(str2_out[1], "def")

    def xtest10(self):
        """test for ticket #74, 'running out of os file descriptors'


        Note: this test takes a very long time, to enable it
        remove the 'X' infront of the test name, to disable it
        add an 'X'.
        Also note: to test this, you best set the ulimit
        to a low number (but not too low), for example
        ulimit -n 400
        """
        for x in range(400):
            instance = ForTestingInterface(self.exefile)
            out, error = instance.echo_float(4.0)
            instance.stop()

    def test11(self):
        instance = ForTestingInterface(self.exefile)
        (output_ints,) = instance.echo_array([4, 5, 6])
        instance.stop()
        self.assertEqual(output_ints[0], 4)
        self.assertEqual(output_ints[1], 5)
        self.assertEqual(output_ints[2], 6)

    def test12(self):
        instance = ForTestingInterface(self.exefile)
        (output_ints, error) = instance.echo_array_with_result([4, 5, 6])
        instance.stop()
        self.assertEqual(output_ints[0], 4)
        self.assertEqual(output_ints[1], 5)
        self.assertEqual(output_ints[2], 6)

        self.assertEqual(error[0], -1)
        self.assertEqual(error[1], -1)
        self.assertEqual(error[2], -1)

    def test13(self):
        instance = ForTesting(self.exefile)
        self.assertRaises(exceptions.AmuseException, instance.echo_int, [-1, -2] | units.m,
            expected_message="Error when calling 'echo_int' of a 'ForTesting', errorcode is -1")
        instance.stop()

    def test14(self):
        instance = ForTesting(self.exefile)
        self.assertRaises(exceptions.CodeException, lambda: instance.echo_int())
        instance.legacy_interface.echo_int.specification.id = -9
        self.assertRaises(exceptions.CodeException, lambda: instance.echo_int(1 | units.m))
        instance.stop()

    def test15(self):
        instance = ForTesting(self.exefile)
        output_ints1, output_ints2 = instance.echo_2_int([1, 2], [3, 4])
        output_ints3, output_ints4 = instance.echo_2_int([1, 2, 3])
        output_ints5, output_ints6 = instance.echo_2_int([5], [0])
        output_ints7, output_ints8 = instance.echo_2_int([5])
        instance.stop()
        self.assertEqual(output_ints1[1], 2)
        self.assertEqual(output_ints2[0], 3)
        self.assertEqual(output_ints2[1], 4)
        for i in range(3):
            self.assertEqual(output_ints3[i], i + 1)
            self.assertEqual(output_ints4[i], 1)
        self.assertEqual(output_ints5[0], 5)
        self.assertEqual(output_ints6[0], 0)
        self.assertEqual(output_ints7[0], 5)
        self.assertEqual(output_ints8[0], 1)

    def test16(self):
        instance = ForTesting(self.exefile)

        # self.assertRaises(exceptions.AmuseException, lambda : instance.echo_int([]))
        instance.stop()

    def test17(self):
        instance = ForTestingInterface(self.exefile)
        (output_ints, error) = instance.echo_inout_array_with_result([4, 5, 6])
        instance.stop()
        self.assertEqual(output_ints[0], 14)
        self.assertEqual(output_ints[1], 15)
        self.assertEqual(output_ints[2], 16)

        self.assertEqual(error[0], 11)
        self.assertEqual(error[1], 11)
        self.assertEqual(error[2], 11)

    def test18(self):
        instance = ForTestingInterface(self.exefile)
        out, error = instance.echo_logical([True, False, True])
        instance.stop()
        self.assertEqual(out, [True, False, True])
        self.assertEqual(error, 0)

    def test19(self):
        instance = ForTestingInterface(self.exefile)
        int_out, error = instance.echo_long_long_int(3935559000370003845)
        instance.stop()
        self.assertEqual(int_out, 3935559000370003845)
        self.assertEqual(error, 0)

    def test20(self):
        path = os.path.abspath(self.get_path_to_results())
        output = os.path.join(path, "output.txt")
        error = os.path.join(path, "error.txt")

        if os.path.exists(output):
            os.remove(output)
        if os.path.exists(error):
            os.remove(error)

        instance = ForTesting(self.exefile, redirect_stderr_file=error, redirect_stdout_file=output, redirection="file")
        instance.print_string("test_string_123")
        instance.print_error_string("error_string_123")
        instance.stop()

        time.sleep(0.2)

        self.assertTrue(os.path.exists(output))
        with open(output, "r") as f:
            content = f.read()
        self.assertTrue("test_string_123" in content.strip())

        self.assertTrue(os.path.exists(error))
        with open(error, "r") as f:
            content = f.read()
        # some times java generates "Picked up _JAVA_OPTIONS" message, so only test:
        self.assertTrue(content.strip().endswith("error_string_123"))

    def test21(self):
        path = os.path.abspath(self.get_path_to_results())
        output = os.path.join(path, "output.txt")
        error = os.path.join(path, "error.txt")

        if os.path.exists(output):
            os.remove(output)
        if os.path.exists(error):
            os.remove(error)

        instance = ForTesting(self.exefile, redirect_stderr_file=output, redirect_stdout_file=output, redirection="file")

        instance.print_string("abcdef")
        instance.print_error_string("&Hfecd")
        instance.stop()

        time.sleep(0.2)

        self.assertTrue(os.path.exists(output))
        with open(output, "r") as f:
            content = f.read()
        # some times java generates "Picked up _JAVA_OPTIONS" message, so only test:
        self.assertTrue("abcdef" in content)
        self.assertTrue("&Hfecd" in content)



from amuse.test import amusetest

from amuse.units import units
from amuse.ic.brokenimf import new_kroupa_mass_distribution


class TestsForIssue777(amusetest.TestCase):
    def test_upper_segment(self):
        "Test if a star in the upper mass segment will get the right mass"
        lower_limit = 1.0 | units.MSun
        upper_limit = 1.0 | units.MSun
        mass = new_kroupa_mass_distribution(
            1,
            mass_min=lower_limit,
            mass_max=upper_limit,
        )
        self.assertEqual(mass[0], 1.0 | units.MSun)

    def test_middle_segment(self):
        "Test if a star in the middle mass segment will get the right mass"
        lower_limit = 0.2 | units.MSun
        upper_limit = 0.2 | units.MSun
        mass = new_kroupa_mass_distribution(
            1,
            mass_min=lower_limit,
            mass_max=upper_limit,
        )
        self.assertEqual(mass[0], 0.2 | units.MSun)

    def test_lower_segment(self):
        "Test if a star in the lower mass segment will get the right mass"
        lower_limit = 0.02 | units.MSun
        upper_limit = 0.02 | units.MSun
        mass = new_kroupa_mass_distribution(
            1,
            mass_min=lower_limit,
            mass_max=upper_limit,
        )
        self.assertEqual(mass[0], 0.02 | units.MSun)
